import logging
import secrets

from urllib.parse import urljoin, urlparse

from bovine import BovineActor
from bovine.activitystreams import Actor as AsActor, factories_for_actor_object
from bovine.activitystreams.utils.property_value import property_value_context
from bovine.crypto import generate_rsa_public_private_key
from bovine.types import Visibility
from sqlalchemy import delete, select
from sqlalchemy.ext.asyncio import AsyncSession


from cattle_grid.database.activity_pub_actor import (
    Actor,
    PublicIdentifier,
    Follower,
    Following,
    ActorStatus,
    PublicIdentifierStatus,
)
from cattle_grid.database.activity_pub import Credential

from .identifiers import (
    determine_preferred_username,
    collect_identifiers_for_actor,
    identifier_in_list_exists,
)
from .helper import endpoints_object_from_actor_id

logger = logging.getLogger(__name__)


class DuplicateIdentifierException(Exception):
    """Raised if an identifier already exists and one tries to create an actor with it"""


def new_url(base_url: str, url_type: str) -> str:
    token = secrets.token_urlsafe(16)
    return urljoin(base_url, f"{url_type}/{token}")


def compute_acct_uri(base_url: str, preferred_username: str):
    """Computes the acct uri

    ```pycon
    >>> compute_acct_uri("http://host.example/somewhere", "alice")
    'acct:alice@host.example'

    ```

    """
    host = urlparse(base_url).hostname

    return f"acct:{preferred_username}@{host}"


async def create_actor(
    session: AsyncSession,
    base_url: str,
    preferred_username: str | None = None,
    identifiers: dict = {},
    profile: dict = {},
):
    """Creates a new actor in the database"""

    public_key, private_key = generate_rsa_public_private_key()
    public_key_name = "legacy-key-1"
    actor_id = new_url(base_url, "actor")

    if preferred_username:
        if "webfinger" in identifiers:
            raise ValueError("webfinger key set in identifiers")
        identifiers = {
            **identifiers,
            "webfinger": compute_acct_uri(base_url, preferred_username),
        }

    if "activitypub_id" not in identifiers:
        identifiers = {**identifiers, "activitypub_id": actor_id}

    identifier_already_exists = await identifier_in_list_exists(
        session, list(identifiers.values())
    )

    if identifier_already_exists:
        raise DuplicateIdentifierException("identifier already exists")

    actor = Actor(
        actor_id=actor_id,
        inbox_uri=new_url(base_url, "inbox"),
        outbox_uri=new_url(base_url, "outbox"),
        following_uri=new_url(base_url, "following"),
        followers_uri=new_url(base_url, "followers"),
        public_key_name=public_key_name,
        public_key=public_key,
        profile={**profile},
        automatically_accept_followers=False,
        status=ActorStatus.active,
    )
    session.add(actor)

    for name, identifier in identifiers.items():
        session.add(
            PublicIdentifier(
                actor=actor,
                name=name,
                identifier=identifier,
                status=PublicIdentifierStatus.verified,
            )
        )

    credential = Credential(
        actor_id=actor_id,
        identifier=f"{actor_id}#{public_key_name}",
        secret=private_key,
    )
    session.add(credential)

    logging.info("Created actor with id '%s'", actor_id)

    await session.commit()
    await session.refresh(actor, attribute_names=["identifiers"])

    return actor


def actor_to_object(actor: Actor) -> dict:
    """Transform the actor to an object

    :params actor:
    :returns:
    """

    sorted_identifiers = collect_identifiers_for_actor(actor)

    preferred_username = determine_preferred_username(
        sorted_identifiers, actor.actor_id
    )
    attachments = actor.profile.get("attachment")
    result = AsActor(
        id=actor.actor_id,
        outbox=actor.outbox_uri,
        inbox=actor.inbox_uri,
        followers=actor.followers_uri,
        following=actor.following_uri,
        public_key=actor.public_key,
        public_key_name=actor.public_key_name,
        preferred_username=preferred_username,
        type=actor.profile.get("type", "Person"),
        name=actor.profile.get("name"),
        summary=actor.profile.get("summary"),
        url=actor.profile.get("url"),
        icon=actor.profile.get("image", actor.profile.get("icon")),
        properties={
            "attachment": attachments,
            "published": actor.created_at.isoformat(),
        },
    ).build(visibility=Visibility.OWNER)

    result["identifiers"] = sorted_identifiers
    result["endpoints"] = endpoints_object_from_actor_id(actor.actor_id)

    if attachments:
        result["@context"].append(property_value_context)

    return result


async def bovine_actor_for_actor_id(
    session: AsyncSession, actor_id: str
) -> BovineActor | None:
    """Uses the information stored in
    [Credential][cattle_grid.database.activity_pub.Credential] to construct a bovine actor

    :params actor_id:
    :returns:
    """
    credential = await session.scalar(
        select(Credential).where(Credential.actor_id == actor_id)
    )

    if credential is None:
        logger.warning("No credential found for %s", actor_id)
        return None

    return BovineActor(
        public_key_url=credential.identifier,
        actor_id=actor_id,
        secret=credential.secret,
    )


def update_for_actor_profile(actor: Actor) -> dict:
    """Creates an update for the Actor"""

    actor_profile = actor_to_object(actor)
    activity_factory, _ = factories_for_actor_object(actor_profile)

    return (
        activity_factory.update(actor_profile, followers=actor_profile["followers"])
        .as_public()
        .build()
    )


def delete_for_actor_profile(actor: Actor) -> dict:
    """Creates a delete activity for the Actor"""

    actor_profile = actor_to_object(actor)
    activity_factory, _ = factories_for_actor_object(actor_profile)

    result = (
        activity_factory.delete(
            actor_profile.get("id"), followers=actor_profile["followers"]
        )
        .as_public()
        .build()
    )

    result["cc"].append(actor_profile["following"])

    return result


async def delete_actor(session: AsyncSession, actor: Actor):
    """Deletes an actor

    :param actor: Actor to be deleted
    """
    await session.execute(
        delete(PublicIdentifier).where(PublicIdentifier.actor_id == actor.id)
    )
    actor.status = ActorStatus.deleted
    await session.commit()


async def remove_from_followers_following(
    session: AsyncSession, actor_id_to_remove: str
):
    """Removes actor_id from all occurring followers and following"""
    await session.execute(
        delete(Follower).where(Follower.follower == actor_id_to_remove)
    )
    await session.execute(
        delete(Following).where(Following.following == actor_id_to_remove)
    )



"""
The exchanges used by cattle_grid are using routing keys
to make processing easier. The cattle_grid gateway takes
these messages and readdresses them with routing keys
based on an user. Here an user can have multiple actors.

Furthermore, convenience methods are provided to manage
users and actors through a HTTP Api. This is in difference
to interacting with the Fediverse, which is done through a
message queue.
"""

from faststream.rabbit import RabbitRouter
from .router import create_router


def create_gateway_router() -> RabbitRouter:
    """Creates a router that moves messages to be routed by user."""
    return create_router()



"""
Implementation of a [HTTP auth backend for rabbitmq](https://github.com/rabbitmq/rabbitmq-server/tree/v3.13.x/deps/rabbitmq_auth_backend_http).

A possible configuration of RabbitMQ is

```conf title="/etc/rabbitmq/conf.d/03_http_auth.conf"
--8<-- "./resources/dev/03_http_auth.conf"
```

Here, we use `auth_backend = internal` for the user
corresponding to the `cattle_grid` processes. As cattle_grid
connects to RabbitMQ on startup, it cannot authenticate
itself.

"""

import logging

from fastapi import APIRouter, Form
from fastapi.responses import PlainTextResponse
from typing import Annotated

from cattle_grid.account.account import account_with_name_password
from cattle_grid.dependencies.fastapi import SqlSession


logger = logging.getLogger(__name__)

rabbit_router = APIRouter(prefix="/rabbitmq", tags=["rabbitmq"])


@rabbit_router.post("/user", response_class=PlainTextResponse)
async def user_auth(
    username: Annotated[str, Form()],
    password: Annotated[str, Form()],
    session: SqlSession,
) -> str:
    """Checks login with username/password"""
    user = await account_with_name_password(session, username, password)

    if not user:
        logger.warning("Failed login to message broker with username '%s'", username)
        return "deny"

    return "allow"


@rabbit_router.post("/vhost", response_class=PlainTextResponse)
async def vhost_auth(
    username: Annotated[str, Form()], vhost: Annotated[str, Form()]
) -> str:
    """Authentication for vhosts, currently only "/" is allowed"""
    if vhost != "/":
        logger.warning("User %s tried to access vhost %s", username, vhost)
        return "deny"
    return "allow"


@rabbit_router.post("/resource", response_class=PlainTextResponse)
async def resource_auth() -> str:
    """Always allowed"""
    return "allow"


def validate_routing_key(username: str, routing_key: str):
    """Rules for the routing key, e.g.

    ```pycon
    >>> validate_routing_key("alice", "send.alice.trigger")
    True

    >>> validate_routing_key("alice", "send.bob.trigger")
    False

    ```
    """
    if routing_key.startswith(f"send.{username}."):
        return True
    if routing_key.startswith(f"receive.{username}."):
        return True
    if routing_key == f"error.{username}":
        return True
    return False


@rabbit_router.post(
    "/topic",
    response_class=PlainTextResponse,
)
async def topic_auth(
    username: Annotated[str, Form()],
    name: Annotated[str, Form()],
    routing_key: Annotated[str, Form()],
) -> str:
    """Checks if topic is allowed. Currently allowed are

    ```
    exchange = "amq.topic"
    ```

    and the routing keys `send.username` and `receive.username`
    """
    if name != "amq.topic":
        logger.warning("User %s tried to access exchange %s", username, name)
        return "deny"
    if not validate_routing_key(username, routing_key):
        logger.warning(
            "User %s tried to subscribe to routing_key %s",
            username,
            routing_key,
        )
        return "deny"

    return "allow"


def app_for_schema():
    from fastapi import FastAPI

    app = FastAPI(title="Authentication API for RabbitMQ")
    app.include_router(rabbit_router)
    return app



"""foreignkeys

Revision ID: 95e4ef3e61b7
Revises: bab34046008c
Create Date: 2025-08-29 13:47:13.119511

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "95e4ef3e61b7"
down_revision: Union[str, Sequence[str], None] = "bab34046008c"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###)
    op.drop_constraint(
        op.f("actorforaccount_account_id_fkey"), "actorforaccount", type_="foreignkey"
    )
    op.create_foreign_key(
        None, "actorforaccount", "account", ["account_id"], ["id"], ondelete="CASCADE"
    )
    op.drop_constraint(
        op.f("actorgroup_actor_id_fkey"), "actorgroup", type_="foreignkey"
    )
    op.create_foreign_key(
        None, "actorgroup", "actorforaccount", ["actor_id"], ["id"], ondelete="CASCADE"
    )
    op.drop_constraint(
        op.f("authenticationtoken_account_id_fkey"),
        "authenticationtoken",
        type_="foreignkey",
    )
    op.create_foreign_key(
        None,
        "authenticationtoken",
        "account",
        ["account_id"],
        ["id"],
        ondelete="CASCADE",
    )
    op.drop_constraint(
        op.f("permission_account_id_fkey"), "permission", type_="foreignkey"
    )
    op.create_foreign_key(
        None, "permission", "account", ["account_id"], ["id"], ondelete="CASCADE"
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "permission", type_="foreignkey")
    op.create_foreign_key(
        op.f("permission_account_id_fkey"),
        "permission",
        "account",
        ["account_id"],
        ["id"],
    )
    op.drop_constraint(None, "authenticationtoken", type_="foreignkey")
    op.create_foreign_key(
        op.f("authenticationtoken_account_id_fkey"),
        "authenticationtoken",
        "account",
        ["account_id"],
        ["id"],
    )
    op.drop_constraint(None, "actorgroup", type_="foreignkey")
    op.create_foreign_key(
        op.f("actorgroup_actor_id_fkey"),
        "actorgroup",
        "actorforaccount",
        ["actor_id"],
        ["id"],
    )
    op.drop_constraint(None, "actorforaccount", type_="foreignkey")
    op.create_foreign_key(
        op.f("actorforaccount_account_id_fkey"),
        "actorforaccount",
        "account",
        ["account_id"],
        ["id"],
    )
    # ### end Alembic commands ###



"""Objects used for the account exchange"""

from typing import List, Dict, Any
from enum import StrEnum, auto
from pydantic import BaseModel, Field, ConfigDict

from .extension import MethodInformationModel


class WithActor(BaseModel):
    """Used as base for messages requiring an actor"""

    actor: str = Field(
        examples=["http://host.example/actor/1"],
        description="""The actor performing the action""",
    )


class FetchMessage(WithActor):
    """Message to fetch an object from the Fediverse"""

    uri: str = Field(
        examples=["http://remote.example/object/1"],
        description="""The resource to fetch""",
    )


class FetchResponse(WithActor):
    """Result of a a fetch request"""

    uri: str = Field(
        examples=["http://remote.example/object/1"],
        description="""The resource that was requested""",
    )

    data: dict | None = Field(description="""The data returned for the object""")


class TriggerMessage(WithActor):
    """Message to trigger something on the ActivityExchange"""

    model_config = ConfigDict(extra="allow")


class CreateActorRequest(BaseModel):
    """Request to create an actor"""

    base_url: str = Field(
        examples=["http://host.example"],
        serialization_alias="baseUrl",
        description="""Base url for the actor, the actor URI will be of the form `{base_url}/actor/{id}`""",
    )
    name: str | None = Field(
        default=None, examples=["alice"], description="""Internal name of the actor"""
    )

    preferred_username: str | None = Field(
        default=None,
        examples=["alice", "bob"],
        description="""
    Add a preferred username. This name will be used in acct:username@domain and supplied to webfinger. Here domain is determine from baseUrl.
    """,
        serialization_alias="preferredUsername",
    )
    profile: Dict[str, Any] = Field(
        default={},
        examples=[{"summary": "A new actor"}],
        description="""
    New profile object for the actor.
    """,
    )
    automatically_accept_followers: bool | None = Field(
        default=None,
        examples=[True, False, None],
        description="""
    Enables setting actors to automatically accept follow requests
    """,
        serialization_alias="automaticallyAcceptFollowers",
    )


class NameAndVersion(BaseModel):
    """Name and version information"""

    name: str = Field(
        examples=["cattle_grid", "CattleDrive"],
        description="""Name of the server or protocol""",
    )

    version: str = Field(
        examples=["3.1.4"],
        description="""Version of the server or protocol""",
    )


class ActorInformation(BaseModel):
    """Information about an actor"""

    id: str = Field(
        description="The URI corresponding to the actor",
        examples=["http://host.example/actor/1"],
    )
    name: str = Field(examples=["Alice"], description="Internal name of the actor")


class InformationResponse(BaseModel):
    """Response for the information request"""

    account_name: str = Field(
        examples=["alice"],
        serialization_alias="accountName",
        description="Name of the account",
    )

    actors: List[ActorInformation] = Field(
        examples=[
            [
                ActorInformation(id="http://host.example/actor/1", name="Alice"),
                ActorInformation(id="http://host.example/actor/2", name="Bob"),
            ]
        ],
        description="Actors of the account on the server",
    )

    base_urls: List[str] = Field(
        examples=[["http://host.example"]],
        serialization_alias="baseUrls",
        description="""The base urls of the server""",
    )

    backend: NameAndVersion = Field(
        examples=[NameAndVersion(name="cattle_grid", version="3.1.4")],
        description="""Name and version of the backend""",
    )

    protocol: NameAndVersion = Field(
        examples=[NameAndVersion(name="CattleDrive", version="3.1.4")],
        description="""Name and version of the protocol being used""",
    )

    method_information: List[MethodInformationModel] = Field(
        default_factory=list,
        examples=[
            [
                MethodInformationModel(
                    routing_key="send_message",
                    module="cattle_grid",
                    description="Send a message as the actor",
                )
            ]
        ],
        serialization_alias="methodInformation",
    )


class EventType(StrEnum):
    """Types of events

    ```pycon
    >>> EventType.incoming.value
    'incoming'

    ```
    """

    incoming = auto()
    outgoing = auto()
    error = auto()


class EventInformation(WithActor):
    """Send on outgoing or incoming events"""

    event_type: EventType = Field(
        examples=[EventType.incoming, EventType.outgoing],
        description="""Type of event
    
incoming means that the contained data was not generated on the actors behalf.

outgoing means that the data is being send out by the actor.""",
    )

    data: Dict[str, Any] = Field(
        examples=[
            {
                "raw": {
                    "@context": "https://www.w3.org/ns/activitystreams",
                    "type": "Create",
                    "actor": "http://host.example/actor/1",
                }
            }
        ],
        description="""The data that was exchanged. We note that this data was processed by the transformers.""",
    )


class ErrorMessage(BaseModel):
    """Message to send an exception"""

    model_config = ConfigDict(extra="allow")

    message: str = Field(description="The exception")

    routing_key: str = Field(
        examples=["send.alice.trigger.method"],
        description="""The routing key of the message that caused the error""",
    )



import argon2
import logging
import re

from sqlalchemy import delete, select
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.config.settings import get_settings
from cattle_grid.database.activity_pub_actor import Actor

from cattle_grid.database.account import (
    Account,
    Permission,
    ActorStatus,
    ActorForAccount,
    ActorGroup,
)

logger = logging.getLogger(__name__)

password_hasher = argon2.PasswordHasher()


class AccountAlreadyExists(Exception):
    pass


class InvalidAccountName(Exception):
    pass


class WrongPassword(Exception):
    pass


async def create_account(
    session: AsyncSession,
    name: str,
    password: str,
    settings=get_settings(),
    permissions: list[str] = [],
    meta_information: dict[str, str] = {},
) -> Account | None:
    """Creates a new account for name and password"""
    if await session.scalar(select(Account).where(Account.name == name)):
        raise AccountAlreadyExists("Account already exists")

    if not re.match(settings.account.allowed_name_regex, name):  # type: ignore
        raise InvalidAccountName("Account name does not match allowed format")

    if name in settings.account.forbidden_names:  # type: ignore
        raise InvalidAccountName("Account name is forbidden")

    account = Account(
        name=name,
        password_hash=password_hasher.hash(password),
        meta_information=meta_information,
    )
    session.add(account)
    await session.commit()

    for permission in permissions:
        await add_permission(session, account, permission)
    return account


async def account_with_name_password(
    session: AsyncSession, name: str, password: str
) -> Account | None:
    """Retrieves account for given name and password"""
    account = await session.scalar(select(Account).where(Account.name == name))
    if account is None:
        return None

    try:
        password_hasher.verify(account.password_hash, password)
    except argon2.exceptions.VerifyMismatchError:
        logger.warning("Got wrong password for %s", name)
        return None

    # Implement rehash?
    # https://argon2-cffi.readthedocs.io/en/stable/howto.html

    return account


async def delete_account(
    session: AsyncSession, name: str, password: str | None = None, force: bool = False
) -> None:
    """Deletes account for given account name and password

    If password is wrong or account does not exist,
    raises a WrongPassword exception"""
    if force:
        account = await session.scalar(select(Account).where(Account.name == name))
    else:
        if not password:
            raise ValueError("Password must be provided if not forcing deletion")
        account = await account_with_name_password(session, name, password)

    if account is None:
        raise WrongPassword(
            "Either the account does not exist or the password is wrong"
        )

    await session.refresh(account, attribute_names=["actors"])

    active_actors = [a for a in account.actors if a.status == ActorStatus.active]
    if len(active_actors) > 0:
        logger.warning(
            "Deleting account with active actors: %s", [a.actor for a in active_actors]
        )

    await session.delete(account)
    await session.commit()


async def add_permission(
    session: AsyncSession, account: Account, permission: str
) -> None:
    """Adds permission to account"""
    session.add(Permission(account=account, name=permission))
    await session.commit()


async def remove_permission(
    session: AsyncSession, account: Account, permission: str
) -> None:
    """Removes permission from account"""
    await session.execute(
        delete(Permission)
        .where(Permission.account == account)
        .where(Permission.name == permission)
    )


def list_permissions(account: Account) -> list[str]:
    """Returns list of permissions for account"""
    return [p.name for p in account.permissions]


async def account_for_actor(session: AsyncSession, actor: str) -> Account | None:
    """Given an actor_id returns the corresponding account or None"""
    actor_for_account = await session.scalar(
        select(ActorForAccount).where(ActorForAccount.actor == actor)
    )

    if not actor_for_account:
        return None

    return actor_for_account.account


async def add_actor_to_account(
    session: AsyncSession, account: Account, actor: Actor, name: str = "added"
) -> None:
    """Adds the actor to the account"""
    session.add(ActorForAccount(account=account, actor=actor.actor_id, name=name))
    await session.commit()


async def add_actor_to_group(
    session: AsyncSession, actor: ActorForAccount, group_name: str
) -> None:
    """Adds the actor to the group"""
    session.add(ActorGroup(actor=actor, name=group_name))
    await session.commit()


async def group_names_for_actor(
    session: AsyncSession, actor: ActorForAccount
) -> list[str]:
    """Returns the group names for an actor"""
    await session.refresh(actor, attribute_names=["groups"])
    return [g.name for g in actor.groups]


async def actor_for_actor_id(
    session: AsyncSession, actor_id: str
) -> ActorForAccount | None:
    """Returns the ActorForAccount for the given actor_id"""
    return await session.scalar(
        select(ActorForAccount).where(ActorForAccount.actor == actor_id)
    )



import logging

from behave import when, given

from uuid import uuid4

from bovine.activitystreams import factories_for_actor_object

from cattle_grid.testing.features import publish_as, send_message_as_actor

logger = logging.getLogger(__name__)


@when('"{alice}" sends "{bob}" a Follow Activity')
async def send_follow(context, alice, bob):
    """Sends a follow Activity. Usage

    ```gherkin
    When "Alice" sends "Bob" a Follow Activity
    ```

    Stores the follow activity in `context.follow_activity`
    """
    alice_actor = context.actors[alice]
    bob_id = context.actors[bob].get("id")
    activity_factory, _ = factories_for_actor_object(alice_actor)

    context.follow_id = "follow:" + str(uuid4())

    context.follow_activity = activity_factory.follow(
        bob_id, id=context.follow_id
    ).build()

    await publish_as(
        context,
        alice,
        "send_message",
        send_message_as_actor(alice_actor, context.follow_activity),
    )


@when('"{actor}" sends an Accept to this Follow Activity')
async def accept_follow_request(context, actor):
    """Checks that Alice received a follow Activity and then
    accepts this follow activity

    ```gherkin
    When "Alice" sends an Accept to this Follow Activity
    ```
    """
    result = await context.connections[actor].next_incoming()
    received_activity = result.get("data")
    if "raw" in received_activity:
        received_activity = received_activity["raw"]

    logger.debug("Got follow request:")
    logger.debug(received_activity)

    assert received_activity["type"] == "Follow"

    follow_id = received_activity["id"]
    to_follow = received_activity["actor"]

    alice = context.actors[actor]
    activity_factory, _ = factories_for_actor_object(alice)

    activity = activity_factory.accept(follow_id, to={to_follow}).build()
    activity["id"] = "accept:" + str(uuid4())

    await publish_as(
        context, actor, "send_message", send_message_as_actor(alice, activity)
    )


@given('"{bob}" follows "{alice}"')
def actor_follows_other(context, bob, alice):
    """Combination of two steps, i.e.

    ```gherkin
    When "Alice" follows "Bob"
    ```

    is the same as

    ```gherkin
    When "Alice" sends "Bob" a Follow Activity
    And "Bob" sends an Accept to this Follow Activity
    ```
    """

    logger.debug("Available conections %s", ", ".join(context.connections.keys()))

    if alice not in context.connections:
        context.execute_steps(
            f"""
        When "{bob}" sends "{alice}" a Follow Activity
    """
        )
    else:
        context.execute_steps(
            f"""
        When "{bob}" sends "{alice}" a Follow Activity
        And "{alice}" sends an Accept to this Follow Activity
    """
        )


@when('"{bob}" sends "{alice}" an Undo Follow Activity')
async def send_undo_follow(context, bob, alice):
    """Sends an Undo Follow activity for the follow activity
    with id stored in `context.follow_activity`.

    Usage:

    ```gherkin
    When "Bob" sends "Alice" an Undo Follow Activity
    ```
    """
    actor = context.actors[bob]
    activity_factory, _ = factories_for_actor_object(actor)

    activity = activity_factory.undo(context.follow_activity).build()
    if isinstance(activity["object"], dict):
        activity["object"] = activity["object"]["id"]

    activity["id"] = "undo:" + str(uuid4())

    await publish_as(
        context, bob, "send_message", send_message_as_actor(actor, activity)
    )


@given('"{alice}" automatically accepts followers')
async def automatically_accept_followers(context, alice):
    """FIXME: Should toggle"""

    actor = context.actors[alice]

    await publish_as(
        context,
        alice,
        "update_actor",
        {"actor": actor.get("id"), "autoFollow": True},
    )


@when('"{alice}" sends "{bob}" a Reject Follow Activity')
async def send_reject_follow(context, alice, bob):
    """Sends an Undo Follow activity for the follow activity
    with id stored in `context.follow_activity`.

    Usage:

    ```gherkin
    When "Alice" sends "Bob" a Reject Follow Activity
    ```
    """
    actor = context.actors[alice]
    activity_factory, _ = factories_for_actor_object(actor)

    activity = activity_factory.reject(context.follow_activity).build()
    if isinstance(activity["object"], dict):
        activity["object"] = activity["object"]["id"]

    activity["id"] = "reject:" + str(uuid4())

    await publish_as(
        context, alice, "send_message", send_message_as_actor(actor, activity)
    )



import os
import asyncio
import aio_pika
import json


async def publish_reporting(routing_key, data):
    mq_host = os.environ.get("CATTLE_GRID_MQ", "rabbitmq")

    connection = await aio_pika.connect_robust(
        f"amqp://guest:guest@{mq_host}/",
    )

    async with connection:
        channel = await connection.channel()

        exchange = await channel.declare_exchange(
            "reporting", aio_pika.ExchangeType.TOPIC
        )

        await exchange.publish(
            aio_pika.Message(body=json.dumps(data).encode()),
            routing_key=routing_key,
        )


def before_step(context, step):
    data = {"name": step.name, "type": step.step_type}
    asyncio.get_event_loop().run_until_complete(publish_reporting("step", data))



import logging

from behave import then, when

from bovine.clients import lookup_uri_with_webfinger

from cattle_grid.testing.features import fetch_request

logger = logging.getLogger(__name__)


@when('One queries webfinger for "{resource}"')
async def query_webfinger(context, resource):
    context.actor_uri, _ = await lookup_uri_with_webfinger(
        context.session, resource, domain="http://abel"
    )


@then('The actor URI of "{username}" is returned')
def result_is_actor_uri(context, username):
    assert context.actor_uri == context.actors[username]["id"]


@then("No actor URI is returned")
def no_actor_uri_result(context):
    assert context.actor_uri is None, (
        f"Actor URI is {context.actor_uri}; should be None"
    )


@when('"{bob}" looks up the actor id of "{alice}"')
async def bob_looks_up_alice(context, bob, alice):
    context.lookup_result = await fetch_request(
        context, bob, context.actors[alice].get("id")
    )


@then("410 gone is returned")
def gone_returned(context):
    logger.info(context.lookup_result)
    assert context.lookup_result["type"] == "Tombstone"

    # FIXME: This is not what I want. Requires bovine
    # to return appropriate information


@then('"{bob}" can fetch this activity')
async def can_fetch_Activity(context, bob):
    activity_id = context.activity.get("id")

    result = await fetch_request(context, bob, activity_id)

    assert result
    assert result.get("type") == "Update"



import logging
from typing import Callable, Dict, Awaitable, List, Annotated, Any
from dataclasses import dataclass, field
from contextlib import AbstractAsyncContextManager

from pydantic import BaseModel
from fast_depends import Depends

from faststream.rabbit import RabbitRouter
from fastapi import APIRouter, Depends as FastAPIDepends

from cattle_grid.model.lookup import Lookup
from cattle_grid.model.extension import MethodInformationModel
from cattle_grid.dependencies.globals import global_container

from .util import skip_method_information

logger = logging.getLogger(__name__)


class BaseConfig(BaseModel): ...


@dataclass
class Extension:
    """Data model for an extension"""

    name: str
    """name of the extension, must be unique"""

    module: str
    """module the extension is defined in, should be set to `__name__`"""

    description: str | None = field(
        default=None,
        metadata={
            "description": "description of the extension. If not set is populated from the docstring of the extension"
        },
    )

    rewrite_group_name: str | None = field(
        default=None,
        metadata={"description": "group name this extensions' rewrite rules apply to"},
    )
    rewrite_rules: dict[str, str] = field(
        default_factory=dict,
        metadata={
            "description": "rewrite rules, for these to take effect the rewrite_group_name has to be set"
        },
    )

    lifespan: Callable[[Any], AbstractAsyncContextManager[None]] | None = None
    """The lifespan function"""

    config_class: Any = BaseConfig
    """Expected configuration class"""

    Config: Any | None = None
    """Annotation to retrieve the configuration, e.g.

    ```python
    @extension.lookup()
    async def lookup(
        lookup: Lookup, config: extension.Config
    ) -> Lookup: ...
    ```
    """
    ConfigFastAPI: Any | None = field(
        default=None,
        metadata={
            "description": """Annotation to retrieve the configuration using FastAPI Depends"""
        },
    )

    configuration: Any | None = None

    activity_router: RabbitRouter | None = field(default=None)
    api_router: APIRouter = field(
        default_factory=APIRouter, metadata={"description": "API router"}
    )
    api_prefix: str = ""

    transformer: Callable[[Dict], Awaitable[Dict]] | None = None
    transformer_inputs: List[str] | None = None
    transformer_outputs: List[str] | None = None

    lookup_method: Callable[[Lookup], Awaitable[Lookup]] | None = None
    lookup_order: int | None = None

    method_information: List[MethodInformationModel] = field(default_factory=list)

    def __post_init__(self):
        def get_config():
            return self.configuration

        self.Config = Annotated[self.config_class, Depends(get_config)]
        self.ConfigFastAPI = Annotated[self.config_class, FastAPIDepends(get_config)]

    def configure(self, config: dict):
        """Configures the extension

        The configuration is validated using the config_class.
        """
        self.configuration = self.config_class.model_validate(config)

    def transform(self, inputs: List[str] = [], outputs: List[str] = []):
        """Allows building the extension via decorator. Usage:

        ```python
        extension = Extension("my extension")

        @extension.transform(inputs=["inputs"], outputs=["outputs"])
        async def transformer(a: dict):
            ...
        ```
        """
        if self.transformer:
            raise ValueError("You should not override an existing transformer")

        def inner(func):
            self.transformer = func
            self.transformer_inputs = inputs
            self.transformer_outputs = outputs

            return func

        return inner

    def subscribe(self, routing_key: str, description: str | None = None):
        """Allows building the extension via decorator.

        ```python
        extension = Extension("my extension")

        @extension.subscribe("routing_key")
        async def subscriber(message: dict): ...
        ```

        Dependency injection is available for the subscriber function.
        """

        def inner(func):
            if description is None:
                function_description = func.__doc__
            else:
                function_description = description

            if not self.activity_router:
                self.activity_router = RabbitRouter()

            self.activity_router.subscriber(
                routing_key, exchange=global_container.exchange, title=routing_key
            )(func)

            if not skip_method_information(routing_key):
                self.method_information.append(
                    MethodInformationModel(
                        module=self.module,
                        routing_key=routing_key,
                        description=function_description,
                    )
                )

            return func

        return inner

    def subscribe_on_account_exchange(self, routing_key: str):
        """Allows building the extension via decorator.

        Dependency injection is available for the subscriber function.
        """

        def inner(func):
            if not self.activity_router:
                self.activity_router = RabbitRouter()
            self.activity_router.subscriber(
                routing_key,
                exchange=global_container.account_exchange,
                title=routing_key,
            )(func)

            return func

        return inner

    def lookup(self):
        """Allows building the extension via decorator.

        ```python
        extension = Extension("my extension")

        @extension.lookup()
        async def lookup(l: Lookup) -> Lookup:
            ...
        ```
        Dependency injection is available for the lookup function.
        """

        def inner(func):
            self.lookup_method = func
            return func

        return inner

    def get(self, path, **kwargs):
        """Allows one to add a get endpoint to the API Router
        of the extension

        Usage:

        ```python
        @extension.get("/path")
        async def get_endpoint():
            pass

        ```
        """

        def inner(func):
            self.api_router.get(path, **kwargs)(func)
            return func

        return inner

    def post(self, path, **kwargs):
        """Allows one to add a post endpoint to the API Router
        of the extension

        Usage:

        ```python
        @extension.post("/path")
        async def post_endpoint():
            pass

        ```
        """

        def inner(func):
            self.api_router.post(path, **kwargs)(func)
            return func

        return inner

    def include_router(self, router: APIRouter, **kwargs):
        """Includes the router as an api router"""

        self.api_router.include_router(router, **kwargs)



"""ActivityPub related functionality"""

import logging
from fastapi import APIRouter, HTTPException
from sqlalchemy import select

from cattle_grid.database.activity_pub_actor import StoredActivity
from cattle_grid.activity_pub.actor.requester import is_valid_requester
from cattle_grid.dependencies.fastapi import SqlSession

from .router import ActivityPubHeaders, ActivityResponse

logger = logging.getLogger(__name__)

ap_router_object = APIRouter()


@ap_router_object.get("/object/{obj_id}", response_class=ActivityResponse)
async def return_object(obj_id, headers: ActivityPubHeaders, session: SqlSession):
    """Returns the stored activities"""

    obj = await session.scalar(
        select(StoredActivity).where(StoredActivity.id == obj_id)
    )
    if obj is None or not isinstance(obj.data, dict):
        raise HTTPException(404)

    if obj.data.get("id") != headers.x_ap_location:
        raise HTTPException(404)

    if headers.x_cattle_grid_requester is None:
        raise HTTPException(401)

    if not await is_valid_requester(
        session, headers.x_cattle_grid_requester, obj.actor, obj.data
    ):
        raise HTTPException(401)

    return obj.data



import importlib
import logging

from fastapi import FastAPI
from faststream.rabbit import RabbitBroker

from cattle_grid.model.extension import MethodInformationModel
from cattle_grid.extensions import Extension

from .transformer import build_transformer
from .lookup import build_lookup
from .lifespan import lifespan_from_extensions

logger = logging.getLogger(__name__)

__all__ = [
    "load_extension",
    "load_extensions",
    "set_globals",
    "lifespan_from_extensions",
    "add_routers_to_broker",
    "add_routes_to_api",
    "collect_method_information",
    "build_transformer",
    "build_lookup",
]


def load_extension(extension_information: dict) -> Extension:
    """Loads a single extension"""
    module_name = extension_information.get("module")

    if module_name is None:
        raise ValueError("module is required")

    module = importlib.import_module(module_name)
    extension = module.extension
    extension.configure(extension_information.get("config", {}))

    if extension.description is None and module.__doc__ is not None:
        extension.description = module.__doc__

    if "lookup_order" in extension_information:
        extension.lookup_order = extension_information["lookup_order"]
    if "api_prefix" in extension_information:
        extension.api_prefix = extension_information["api_prefix"]

    return extension


def load_extensions(settings) -> list[Extension]:
    """Loads the extensions from settings"""

    extensions = [
        load_extension(extension_information)
        for extension_information in settings.extensions
    ]

    logger.info("Loaded extensions: %s", ", ".join(f"'{e.name}'" for e in extensions))

    return extensions


def set_globals(extensions: list[Extension]):
    """Sets global variables in cattle_grid.dependencies"""
    from cattle_grid.dependencies.globals import global_container

    global_container.transformer = build_transformer(extensions)
    global_container.lookup = build_lookup(extensions)

    for extension in extensions:
        if extension.rewrite_group_name:
            if not global_container._rewrite_rules:
                raise Exception("Rewrite rules not loaded yet")
            global_container._rewrite_rules.add_rules(
                extension.rewrite_group_name, extension.rewrite_rules
            )


def add_routers_to_broker(broker: RabbitBroker, extensions: list[Extension]):
    """Adds the routers to the broker"""

    for extension in extensions:
        if extension.activity_router:
            broker.include_router(extension.activity_router)


def add_routes_to_api(app: FastAPI, extensions: list[Extension]):
    """Adds the routes to the api"""
    for extension in extensions:
        if extension.api_router:
            if extension.api_prefix:
                app.include_router(extension.api_router, prefix=extension.api_prefix)


def collect_method_information(
    extensions: list[Extension],
) -> list[MethodInformationModel]:
    """Collects the method information from the extensions"""
    return sum((extension.method_information for extension in extensions), [])



from unittest.mock import MagicMock
from faststream.rabbit import RabbitBroker, TestRabbitBroker

from cattle_grid.model.common import WithActor

from cattle_grid.testing.fixtures import *  # noqa

from .processing import MessageActor, ActorProfile, FactoriesForActor


async def test_message_actor(actor_for_test):
    broker = RabbitBroker()
    subscriber = MagicMock()

    @broker.subscriber("test")
    async def test_subscriber(actor: MessageActor):
        subscriber(actor.actor_id)

    async with TestRabbitBroker(broker) as br:
        await br.publish(WithActor(actor=actor_for_test.actor_id), "test")

    subscriber.assert_called_once_with(actor_for_test.actor_id)


async def test_actor_profile(actor_for_test):
    broker = RabbitBroker()
    subscriber = MagicMock()

    @broker.subscriber("test")
    async def test_subscriber(actor: ActorProfile):
        subscriber(actor)

    async with TestRabbitBroker(broker) as br:
        await br.publish(WithActor(actor=actor_for_test.actor_id), "test")

    subscriber.assert_called_once()

    (profile,) = subscriber.call_args[0]

    assert isinstance(profile, dict)
    assert profile["id"] == actor_for_test.actor_id


async def test_factories(actor_for_test):
    broker = RabbitBroker()
    subscriber = MagicMock()

    @broker.subscriber("test")
    async def test_subscriber(factories: FactoriesForActor):
        subscriber(factories)

    async with TestRabbitBroker(broker) as br:
        await br.publish(WithActor(actor=actor_for_test.actor_id), "test")

    subscriber.assert_called_once()

    (factories,) = subscriber.call_args[0]

    assert len(factories) == 2



import logging

from faststream import Context
from faststream.rabbit import RabbitRouter
from sqlalchemy import select, delete

from bovine.activitystreams.utils import id_for_object

from cattle_grid.activity_pub.actor import remove_from_followers_following
from cattle_grid.database.activity_pub_actor import Following, Follower
from cattle_grid.model import ActivityMessage
from cattle_grid.model.processing import StoreActivityMessage
from cattle_grid.dependencies.globals import global_container
from cattle_grid.dependencies.processing import MessageActor, FactoriesForActor
from cattle_grid.dependencies import CommittingSession
from .util import actor_deletes_themselves


logger = logging.getLogger(__name__)


async def incoming_follow_request(
    message: ActivityMessage,
    actor: MessageActor,
    factories: FactoriesForActor,
    session: CommittingSession,
    broker=Context(),
):
    """
    For an incoming Follow request an entry in the Follower table is created
    with having `accepted` set to False.

    If the actor automatically accepts followers, the actor sends Accept activity
    to the actor requesting to follow it.
    """
    follow_request = message.data
    to_follow = follow_request.get("object")
    if isinstance(to_follow, dict):
        to_follow = to_follow.get("id")

    if to_follow is None or to_follow != actor.actor_id:
        return

    request_id = follow_request.get("id")
    follower = follow_request.get("actor")

    await session.merge(
        Follower(
            actor=actor,
            follower=follower,
            request=request_id,
            accepted=False,
        )
    )

    if actor.automatically_accept_followers:
        accept = factories[0].accept(follow_request).build()

        await broker.publish(
            StoreActivityMessage(actor=actor.actor_id, data=accept),
            routing_key="store_activity",
            exchange=global_container.internal_exchange,
        )
        logger.info(
            "Got follow request from %s with id %s (auto accepted)",
            follower,
            request_id,
        )

    else:
        logger.info("Got follow request from %s with id %s", follower, request_id)


async def incoming_accept_activity(
    message: ActivityMessage,
    actor: MessageActor,
    session: CommittingSession,
    broker=Context(),
):
    """Handles an incoming Accept activity"""
    accept_request = message.data
    request_being_accepted = id_for_object(accept_request.get("object"))

    following = await session.scalar(
        select(Following).where(Following.request == request_being_accepted)
    )

    if not following:
        logger.warning("Follow request with id '%s' not found", request_being_accepted)
        return

    following.accepted = True
    logger.info("Processed follow request %s (following)", request_being_accepted)


async def incoming_undo_activity(
    message: ActivityMessage,
    actor: MessageActor,
    session: CommittingSession,
    broker=Context(),
):
    """Handles an incoming Undo activity"""

    accept_request = message.data
    request_being_undone = id_for_object(accept_request.get("object"))

    await session.execute(
        delete(Follower).where(Follower.request == request_being_undone)
    )


async def incoming_reject_activity(
    message: ActivityMessage,
    actor: MessageActor,
    session: CommittingSession,
    broker=Context(),
):
    """Handles an incoming Reject activity"""
    reject_request = message.data
    request_being_rejected = reject_request.get("object")

    await session.execute(
        delete(Following).where(Following.request == request_being_rejected)
    )


async def incoming_delete_activity(
    message: ActivityMessage, session: CommittingSession, broker=Context()
):
    if actor_deletes_themselves(message.data):
        actor_id = id_for_object(message.data.get("actor"))
        if actor_id is None:
            return
        logger.info("Got self delete for actor with id '%s'", actor_id)

        await remove_from_followers_following(session, actor_id)


async def incoming_block_activity(
    message: ActivityMessage,
    actor: MessageActor,
    session: CommittingSession,
    broker=Context(),
):
    """Handles an incoming Block activity"""
    current_actor_id = message.data.get("object")
    if current_actor_id != actor.actor_id:
        logger.warning("Mismatch of actor and target of block")
        return

    actor_blocking = message.data.get("actor")

    await session.execute(
        delete(Following).where(Following.following == actor_blocking)
    )


def create_incoming_router(exchange=None):
    router = RabbitRouter()

    if exchange is None:
        exchange = global_container.internal_exchange

    for routing_key, coroutine in [
        ("incoming.Follow", incoming_follow_request),
        ("incoming.Accept", incoming_accept_activity),
        ("incoming.Undo", incoming_undo_activity),
        ("incoming.Delete", incoming_delete_activity),
        ("incoming.Block", incoming_block_activity),
        ("incoming.Reject", incoming_reject_activity),
    ]:
        router.subscriber(
            routing_key,
            exchange=exchange,
            title=routing_key,
        )(coroutine)

    return router



import pytest

from .identifiers import is_identifier_part_of_base_urls, is_identifier_for_a_base_url


@pytest.mark.parametrize(
    "identifier, base_urls, expected",
    [
        ("acct:test@one.test", ["http://one.test"], True),
        ("acct:test@two.test", ["http://one.test"], False),
    ],
)
def test_is_identifier_part_of_base_urls(identifier, base_urls, expected):
    assert is_identifier_part_of_base_urls(identifier, base_urls) == expected


@pytest.mark.parametrize(
    "identifier, base_urls, expected",
    [
        ("acct:test@one.test", ["http://one.test"], True),
        ("acct:test@two.test", ["http://one.test"], False),
        ("http://one.test/actor", ["http://one.test"], True),
        ("http://two.test/actor", ["http://one.test"], False),
        ("https://one.test/actor", ["http://one.test"], True),
        ("https://two.test/actor", ["http://one.test"], False),
    ],
)
def test_is_identifier_for_a_base_url(identifier, base_urls, expected):
    assert is_identifier_for_a_base_url(identifier, base_urls) == expected



from unittest.mock import AsyncMock, MagicMock


from .message_handlers import send_message


async def test_send_message():
    msg = MagicMock(actor="http://actor.test", data={})

    await send_message(msg, correlation_id="123", broker=AsyncMock())


async def test_send_message_context_serialization():
    activity = dict(
        id="http://local.test/id1",
        type="Create",
        actor="http://local.test",
        to=["http://remote.test"],
    )
    activity["@context"] = "https://www.w3.org/ns/activitystreams"
    msg = MagicMock(actor="http://actor.test", data=activity)

    broker = AsyncMock()

    await send_message(msg, correlation_id="123", broker=broker)

    broker.publish.assert_awaited()

    args = broker.publish.call_args[0][0]

    resulting_activity = args.data
    assert "@context" in resulting_activity



from uuid import uuid4
from behave import when, given

from bovine.activitystreams import factories_for_actor_object


from cattle_grid.testing.features import (
    publish_as,
    id_generator_for_actor,
    send_message_as_actor,
)


@given('"{alice}" blocks "{bob}"')  # type: ignore
@when('"{alice}" blocks "{bob}"')  # type: ignore
async def send_block(context, alice, bob):
    """
    ```gherkin
    When "Alice" blocks "Bob"
    ```

    The id of the block is stored in `context.block_id`
    """

    actor = context.actors[alice]
    activity_factory, object_factory = factories_for_actor_object(
        actor, id_generator=id_generator_for_actor(actor)
    )

    bob_id = context.actors[bob].get("id")

    activity = activity_factory.custom(type="Block", object=bob_id, to={bob_id}).build()

    context.block_id = activity.get("id")

    await publish_as(
        context,
        alice,
        "send_message",
        send_message_as_actor(actor, activity),
    )


@when('"{bob}" unblocks "{alice}"')  # type: ignore
async def unblock(context, bob, alice):
    """Sends an Undo Block activity for
    the id stored in `context.block_id`.

    Usage:

    ```gherkin
    When "Bob" unblocks "Alice"
    ```
    """
    actor = context.actors[bob]
    activity_factory, _ = factories_for_actor_object(actor)
    alice_id = context.actors[alice].get("id")

    activity = activity_factory.custom(
        object=context.block_id, to={alice_id}, type="Undo"
    ).build()
    activity["id"] = "undo:" + str(uuid4())

    await publish_as(
        context,
        bob,
        "send_message",
        send_message_as_actor(actor, activity),
    )



from cattle_grid.extensions import Extension
from . import load_extension, build_transformer


def test_load_extension():
    result = load_extension({"module": "cattle_grid.extensions.examples.recipients"})

    assert isinstance(result, Extension)


def test_load_extension_set_options():
    result = load_extension(
        {"module": "cattle_grid.extensions.examples.recipients", "lookup_order": 10}
    )

    assert isinstance(result, Extension)
    assert result.lookup_order == 10


async def test_build_transformer():
    extensions = [
        load_extension({"module": "cattle_grid.extensions.examples.recipients"})
    ]

    transformer = build_transformer(extensions)

    result = await transformer({"raw": {}})

    assert "recipients" in result



import pytest
from sqlalchemy import func

from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.database.account import Account, ActorForAccount
from .account import (
    account_with_name_password,
    create_account,
    delete_account,
    AccountAlreadyExists,
    InvalidAccountName,
    WrongPassword,
    add_permission,
    list_permissions,
    remove_permission,
    account_for_actor,
    add_actor_to_group,
    group_names_for_actor,
    actor_for_actor_id,
)


async def test_wrong_password(sql_session_for_test):
    sql_session_for_test.add(
        Account(
            name="name",
            password_hash="$argon2id$v=19$m=65536,t=3,p=4$MIIRqgvgQbgj220jfp0MPA$YfwJSVjtjSU0zzV/P3S9nnQ/USre2wvJMjfCIjrTQbg",
        )
    )
    await sql_session_for_test.commit()

    result = await account_with_name_password(sql_session_for_test, "name", "pass")

    assert result is None


async def test_create_and_then_get(sql_session_for_test):
    name = "user"
    password = "pass"

    await create_account(sql_session_for_test, name, password)

    result = await account_with_name_password(sql_session_for_test, name, password)

    assert result
    assert result.name == name


async def test_create_duplicate_raises_exception(sql_session_for_test):
    name = "user"
    password = "pass"

    await create_account(sql_session_for_test, name, password)

    with pytest.raises(AccountAlreadyExists):
        await create_account(sql_session_for_test, name, password)


@pytest.mark.parametrize(
    "name", ["", "abcdefghijklmnopqrstuvwxyz", "first.second", "admin"]
)
async def test_create_name_raises_exception(sql_session_for_test, name):
    with pytest.raises(InvalidAccountName):
        await create_account(sql_session_for_test, name, "pass")


async def test_create_and_then_delete_wrong_password(sql_session_for_test):
    name = "user"
    password = "pass"

    await create_account(sql_session_for_test, name, password)

    assert 1 == await sql_session_for_test.scalar(func.count(Account.id))

    with pytest.raises(WrongPassword):
        await delete_account(sql_session_for_test, name, "wrong")


async def test_create_and_then_delete(sql_session_for_test):
    name = "user"
    password = "pass"

    await create_account(sql_session_for_test, name, password)

    assert 1 == await sql_session_for_test.scalar(func.count(Account.id))

    await delete_account(sql_session_for_test, name, password)

    assert 0 == await sql_session_for_test.scalar(func.count(Account.id))


async def test_add_permission(sql_session_for_test):
    name = "user"
    password = "pass"

    account = await create_account(sql_session_for_test, name, password)
    assert account
    await add_permission(sql_session_for_test, account, "admin")
    await add_permission(sql_session_for_test, account, "test")

    await sql_session_for_test.refresh(account, attribute_names=["permissions"])

    assert set(list_permissions(account)) == {"admin", "test"}


async def test_remove_permission(sql_session_for_test):
    name = "user"
    password = "pass"

    account = await create_account(sql_session_for_test, name, password)
    assert account
    await add_permission(sql_session_for_test, account, "admin")
    await add_permission(sql_session_for_test, account, "test")

    await sql_session_for_test.refresh(account, attribute_names=["permissions"])

    assert set(list_permissions(account)) == {"admin", "test"}

    await remove_permission(sql_session_for_test, account, "admin")
    await sql_session_for_test.refresh(account, attribute_names=["permissions"])
    assert set(list_permissions(account)) == {"test"}


async def test_account_for_actor_not_found(sql_session_for_test):
    account_or_none = await account_for_actor(
        sql_session_for_test, "http://actor.example"
    )

    assert account_or_none is None


async def test_account_for_actor(sql_session_for_test):
    account = await create_account(sql_session_for_test, "name", "password")
    actor_id = "http://actor.example"

    sql_session_for_test.add(ActorForAccount(account=account, actor=actor_id))
    await sql_session_for_test.commit()

    result = await account_for_actor(sql_session_for_test, actor_id)

    assert result == account

    result_actor = await actor_for_actor_id(sql_session_for_test, actor_id)
    assert result_actor
    assert result_actor.actor == actor_id


async def test_account_groups(sql_session_for_test):
    account = await create_account(sql_session_for_test, "name", "password")
    actor_id = "http://actor.example"

    actor_for_account = ActorForAccount(account=account, actor=actor_id)
    sql_session_for_test.add(actor_for_account)
    await sql_session_for_test.commit()

    await add_actor_to_group(sql_session_for_test, actor_for_account, "group1")
    await add_actor_to_group(sql_session_for_test, actor_for_account, "group2")

    result = await group_names_for_actor(sql_session_for_test, actor_for_account)

    assert set(result) == {"group1", "group2"}



import datetime

from sqlalchemy import func, JSON
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from sqlalchemy_utils.types import UUIDType


class Base(AsyncAttrs, DeclarativeBase):
    """Base model"""

    pass


class StoredActivity(Base):
    """Stored activity in the database"""

    __tablename__ = "simple_storage_stored_activity"
    """name of the table"""

    id: Mapped[bytes] = mapped_column(UUIDType(binary=True), primary_key=True)
    """The id (uuid as bytes)"""
    data: Mapped[dict] = mapped_column(JSON)
    """The activity as JSON"""
    actor: Mapped[str] = mapped_column()
    """The actor that created the activity"""
    create_date: Mapped[datetime.datetime] = mapped_column(server_default=func.now())
    """The create timestamp"""


class StoredObject(Base):
    """Stored object in the database"""

    __tablename__ = "simple_storage_stored_object"
    """name of the table"""

    id: Mapped[bytes] = mapped_column(UUIDType(binary=True), primary_key=True)
    """The id (uuid as bytes)"""
    data: Mapped[dict] = mapped_column(JSON)
    """The object as JSON"""
    actor: Mapped[str] = mapped_column()
    """The actor that created the object"""
    create_date: Mapped[datetime.datetime] = mapped_column(server_default=func.now())
    """The create timestamp"""



from urllib.parse import urljoin


def shared_inbox_from_actor_id(actor_id: str) -> str:
    """Returns the shared inbox of the actor identified by actor_id

    ```pycon
    >>> shared_inbox_from_actor_id("http://host.test/actor/someId")
    'http://host.test/shared_inbox'

    ```
    """

    return urljoin(actor_id, "/shared_inbox")


def endpoints_object_from_actor_id(actor_id: str) -> dict:
    """Returns the endpoints object of the actor identified by actor_id

    ```pycon
    >>> endpoints_object_from_actor_id("http://host.test/actor/someId")
    {'sharedInbox': 'http://host.test/shared_inbox'}

    ```
    """
    return {"sharedInbox": shared_inbox_from_actor_id(actor_id)}



from behave import then, when, given
from cattle_grid.testing.features import fetch_request


@when('"{actor}" retrieves the object with the actor id of "{target}"')  # pyright: ignore[reportCallIssue]
async def retrieve_object(context, actor, target):
    context.result = await fetch_request(
        context,
        actor,
        context.actors[target].get("id"),
    )


@when('"{actor}" retrieves the object with id "{uri}"')  # pyright: ignore[reportCallIssue]
async def retrieve_object_by_uri(context, actor, uri):
    context.result = await fetch_request(
        context,
        actor,
        uri,
    )


@then('The retrieved object is the profile of "{username}"')  # pyright: ignore[reportCallIssue]
def check_result_is_actor(context, username):
    assert context.result.get("type") == "Person"
    assert context.result.get("id") == context.actors[username].get("id")


@then("no result is returned")  # type: ignore
def no_result(context):
    assert context.result is None


@given('"Alice" has messaged "Bob"')  # type: ignore
def step_impl(context):
    context.execute_steps(
        """
        When "Alice" sends "Bob" a message saying "Hello Bob"
        Then "Bob" receives a message saying "Hello Bob"
    """
    )



from typing import Annotated
from fast_depends import Depends

from cattle_grid.model.lookup import Lookup


async def one(a: Lookup) -> Lookup:
    return Lookup(uri=a.uri, result={"id": a.uri, "method": "one"}, actor=a.actor)


async def skipped(a: Lookup) -> Lookup:
    return a


async def changed(a: Lookup) -> Lookup:
    return Lookup(uri="http://changed.test", actor=a.actor)


async def dependency() -> str:
    return "dependency"


async def with_dependency(a: Lookup, b: Annotated[str, Depends(dependency)]) -> Lookup:
    return Lookup(uri=a.uri, result={"from": b}, actor=a.actor)



import aiohttp

from typing import Annotated, Callable, Awaitable, Dict, List
from dynaconf import Dynaconf
from faststream.rabbit import RabbitBroker, RabbitExchange

from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker, AsyncSession

from fastapi import Depends

from cattle_grid.model.extension import MethodInformationModel
from .globals import (
    get_engine,
    get_transformer,
    get_method_information,
    global_container,
)

SqlAsyncEngine = Annotated[AsyncEngine, Depends(get_engine)]
"""Returns the SqlAlchemy AsyncEngine"""

Transformer = Annotated[Callable[[Dict], Awaitable[Dict]], Depends(get_transformer)]
"""The transformer loaded from extensions"""

Broker = Annotated[RabbitBroker, Depends(global_container.get_broker)]
"""The RabbitMQ broker"""
InternalExchange = Annotated[
    RabbitExchange, Depends(global_container.get_internal_exchange)
]

ActivityExchange = Annotated[RabbitExchange, Depends(global_container.get_exchange)]
"""The Activity Exchange"""


MethodInformation = Annotated[
    List[MethodInformationModel], Depends(get_method_information)
]
"""Returns the information about the methods that are a part of the exchange"""


async def with_fast_api_session(sql_engine: SqlAsyncEngine):
    async with async_sessionmaker(sql_engine)() as session:
        yield session


SqlSession = Annotated[AsyncSession, Depends(with_fast_api_session)]
"""Session annotation to be used with FastAPI"""


async def with_committing_sql_session(session: SqlSession):
    yield session
    await session.commit()


CommittingSqlSession = Annotated[AsyncSession, Depends(with_committing_sql_session)]
"""Session annotation to be used with FastAPI. A commit is performed, after processing the request"""


Config = Annotated[Dynaconf, Depends(global_container.get_config)]
"""Returns the configuration"""


async def get_client_session():
    yield global_container.session


ClientSession = Annotated[aiohttp.ClientSession, Depends(get_client_session)]
"""The [aiohttp.ClientSession][] used by the application"""



from copy import deepcopy
from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.model.exchange_update_actor import UpdateUrlAction, UpdateActionType

from .urls import add_url, remove_url


def test_handle_add_url(actor_for_test):
    url = "http://local.test/html/page"
    add_url(
        actor_for_test,
        UpdateUrlAction(
            action=UpdateActionType.add_url, url=url, media_type="text/html"
        ),
    )

    urls = actor_for_test.profile.get("url", [])

    assert len(urls) >= 1

    assert urls[-1] == {"type": "Link", "href": url, "mediaType": "text/html"}


def test_handle_add_url_is_idempotent(actor_for_test):
    url = "http://local.test/html/page"
    add_url(
        actor_for_test,
        UpdateUrlAction(
            action=UpdateActionType.add_url, url=url, media_type="text/html"
        ),
    )

    urls = deepcopy(actor_for_test.profile.get("url", []))

    add_url(
        actor_for_test,
        UpdateUrlAction(
            action=UpdateActionType.add_url, url=url, media_type="text/html"
        ),
    )

    urls_again = actor_for_test.profile.get("url", [])

    assert urls == urls_again


def test_handle_remove_url(actor_for_test):
    url = "http://local.test/html/page"
    add_url(
        actor_for_test,
        UpdateUrlAction(
            action=UpdateActionType.add_url, url=url, media_type="text/html"
        ),
    )
    remove_url(
        actor_for_test,
        UpdateUrlAction(action=UpdateActionType.remove_url, url=url),
    )

    urls = actor_for_test.profile.get("url", [])

    assert len(urls) == 0



from fastapi import FastAPI
from contextlib import asynccontextmanager

from cattle_grid.dependencies.globals import global_container
from cattle_grid.database.auth import Base
from cattle_grid.version import __version__
from .router import auth_router

__all__ = ["auth_router"]


def create_app():
    """Allows running just the auth endpoint"""

    @asynccontextmanager
    async def lifespan(app):
        async with global_container.alchemy_database() as engine:
            async with engine.begin() as conn:
                await conn.run_sync(Base.metadata.create_all)

            yield

    app = FastAPI(
        lifespan=lifespan,
        title="cattle_grid.auth",
        description="""Authorization server for Fediverse applications. It basically checks HTTP Signatures for you.""",
        version=__version__,
    )

    app.include_router(auth_router)

    return app



from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.dependencies.globals import global_container

from .account import list_permissions
from cattle_grid.database.account import Account


async def allowed_base_urls(session: AsyncSession, account: Account) -> list[str]:
    """Returns the set of base_urls the account
    is allowed to use to create an actor"""
    settings = global_container.config
    await session.refresh(account, attribute_names=["permissions"])

    permissions = list_permissions(account)

    if "admin" in permissions:
        frontend_settings = settings.get("frontend", {})  # type:ignore

        return frontend_settings.get("base_urls", [])

    permission_settings = settings.get("permissions", {})  # type:ignore

    return sum(
        (permission_settings.get(p, {}).get("base_urls", []) for p in permissions),
        [],
    )


async def can_create_actor_at_base_url(
    session: AsyncSession, account: Account, base_url
) -> bool:
    """Checks if the account is allowed to create an actor
    at the base url"""
    allowed_urls = await allowed_base_urls(session, account)

    return base_url in allowed_urls



"""ActivityPub related functionality"""

import logging
import json
from fastapi import APIRouter, HTTPException, Header, Request
from typing import Annotated
from bovine.crypto.digest import validate_digest
from sqlalchemy import select

from cattle_grid.activity_pub.server.validate import validate_request
from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.activity_pub.enqueuer import (
    enqueue_from_inbox,
    enqueue_from_shared_inbox,
)

from cattle_grid.dependencies.fastapi import Broker, InternalExchange, SqlSession
from cattle_grid.tools.fastapi import APHeaders

logger = logging.getLogger(__name__)

ap_router_inbox = APIRouter()


class APHeadersWithDigest(APHeaders):
    """The addition of digest headers"""

    digest: str | None = None
    """Legacy digest"""
    content_digest: str | None = None
    """Digest according to [RFC 9530 Digest Fields](https://www.rfc-editor.org/rfc/rfc9530.html)"""


def validate_digest_header(headers: APHeadersWithDigest, data: bytes):
    digest_headers = {}
    if headers.digest:
        digest_headers["digest"] = headers.digest
    if headers.content_digest:
        digest_headers["content-digest"] = headers.content_digest

    if not validate_digest(digest_headers, data):
        raise HTTPException(400)


def parse_body(raw_body: bytes) -> dict:
    data = json.loads(raw_body)
    if not isinstance(data, dict):
        logger.info("Could not parse request body")
        logger.debug(data)
        raise HTTPException(422)

    return data


@ap_router_inbox.post("/inbox/{id_str}", status_code=202)
async def inbox(
    id_str,
    request: Request,
    headers: Annotated[APHeadersWithDigest, Header()],
    broker: Broker,
    exchange: InternalExchange,
    session: SqlSession,
):
    """Processes an inbox message"""
    logger.info("Got incoming request")
    actor = await session.scalar(
        select(Actor).where(Actor.inbox_uri == headers.x_ap_location)
    )
    actor = await validate_request(session, actor, headers.x_cattle_grid_requester)

    try:
        raw_body = await request.body()
        validate_digest_header(headers, raw_body)
        data = parse_body(raw_body)

        request_actor = data.get("actor")

        if request_actor != headers.x_cattle_grid_requester:
            raise HTTPException(401)

        await enqueue_from_inbox(broker, exchange, actor.actor_id, data)

        return ""

    except Exception as e:
        if isinstance(e, HTTPException):
            raise e
        logger.error("Processing post request failed with %s", e)
        logger.exception(e)

        raise HTTPException(422)


@ap_router_inbox.post("/shared_inbox", status_code=202)
async def shared_inbox(
    request: Request,
    headers: Annotated[APHeadersWithDigest, Header()],
    broker: Broker,
    exchange: InternalExchange,
):
    try:
        raw_body = await request.body()
        validate_digest_header(headers, raw_body)
        data = parse_body(raw_body)

        request_actor = data.get("actor")

        if request_actor != headers.x_cattle_grid_requester:
            raise HTTPException(401)

        await enqueue_from_shared_inbox(broker, exchange, data)

        return ""

    except Exception as e:
        if isinstance(e, HTTPException):
            raise e
        logger.error("Processing post request failed with %s", e)
        logger.exception(e)

        raise HTTPException(422)



import pytest
from datetime import datetime, timezone
from bovine import BovineActor
from sqlalchemy import select
from sqlalchemy.sql import func

from cattle_grid.database.activity_pub import Credential
from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.database.activity_pub_actor import (
    Actor,
    PublicIdentifier,
    Follower,
    Following,
    ActorStatus,
)


from . import (
    create_actor,
    actor_to_object,
    compute_acct_uri,
    bovine_actor_for_actor_id,
    update_for_actor_profile,
    delete_for_actor_profile,
    delete_actor,
    remove_from_followers_following,
    DuplicateIdentifierException,
)


async def test_create_actor(sql_session_for_test):
    actor = await create_actor(sql_session_for_test, "http://localhost/ap/")

    await sql_session_for_test.refresh(actor)

    assert actor.actor_id.startswith("http://localhost/ap/actor/")

    assert 1 == await sql_session_for_test.scalar(func.count(Actor.id))


async def test_create_then_delete_actor(sql_session_for_test):
    actor = await create_actor(
        sql_session_for_test, "http://localhost/ap/", preferred_username="me"
    )
    await delete_actor(sql_session_for_test, actor)

    assert 1 == await sql_session_for_test.scalar(func.count(Actor.id))
    await sql_session_for_test.refresh(actor)

    assert actor.status == ActorStatus.deleted

    assert 0 == await sql_session_for_test.scalar(func.count(PublicIdentifier.id))
    assert 1 == await sql_session_for_test.scalar(select(func.count(Credential.id)))


async def test_actor_to_object(sql_session_for_test):
    actor = await create_actor(sql_session_for_test, "http://localhost/ap/")

    actor.created_at = datetime(2020, 4, 7, 12, 56, 12, tzinfo=timezone.utc)

    obj = actor_to_object(actor)

    assert obj["type"] == "Person"

    assert obj.get("followers")
    assert obj["published"] == "2020-04-07T12:56:12+00:00"


async def test_update_for_actor_profile(sql_session_for_test):
    actor = await create_actor(sql_session_for_test, "http://localhost/ap/")

    activity = update_for_actor_profile(actor)

    assert activity["type"] == "Update"
    obj = activity["object"]

    assert obj["type"] == "Person"
    assert activity["cc"]


async def test_actor_to_object_with_profile(sql_session_for_test):
    actor = await create_actor(
        sql_session_for_test,
        "http://localhost/ap/",
        profile={"name": "Alice Newton", "type": "Application"},
    )

    obj = actor_to_object(actor)

    assert obj["type"] == "Application"

    assert obj.get("name") == "Alice Newton"


async def test_actor_to_object_with_image(sql_session_for_test):
    actor = await create_actor(
        sql_session_for_test,
        "http://localhost/ap/",
        profile={"image": {"type": "Image"}},
    )

    obj = actor_to_object(actor)

    assert obj["icon"] == {"type": "Image"}


async def test_create_actor_with_identifiers(sql_session_for_test):
    identifier = "acct:you@localhost"
    await create_actor(
        sql_session_for_test,
        "http://localhost/ap/",
        identifiers={"webfinger": identifier},
    )

    pi = await sql_session_for_test.scalar(
        select(PublicIdentifier).where(PublicIdentifier.identifier == identifier)
    )

    assert pi

    assert pi.name == "webfinger"
    assert pi.identifier == identifier


async def test_create_actor_with_preferred_username(sql_session_for_test):
    identifier = "acct:me@localhost"
    actor = await create_actor(
        sql_session_for_test, "http://localhost/ap/", preferred_username="me"
    )

    pi = await sql_session_for_test.scalar(
        select(PublicIdentifier).where(PublicIdentifier.identifier == identifier)
    )

    assert pi

    assert pi.name == "webfinger"
    assert pi.identifier == identifier

    profile = actor_to_object(actor)
    assert profile["preferredUsername"] == "me"


def test_compute_webfinger():
    webfinger = compute_acct_uri("http://localhost/ap", "me")

    assert webfinger == "acct:me@localhost"


async def test_get_bovine_actor_not_found(sql_session_for_test):
    bovine_actor = await bovine_actor_for_actor_id(
        sql_session_for_test, "http://nothing.here/actor"
    )

    assert bovine_actor is None


async def test_get_bovine_actor(sql_session_for_test):
    actor = await create_actor(sql_session_for_test, "http://localhost/ap/")

    bovine_actor = await bovine_actor_for_actor_id(sql_session_for_test, actor.actor_id)

    assert isinstance(bovine_actor, BovineActor)


async def test_delete_for_actor_profile(sql_session_for_test):
    actor = await create_actor(sql_session_for_test, "http://localhost/ap/")

    activity = delete_for_actor_profile(actor)

    assert activity["type"] == "Delete"

    assert activity["object"] == actor.actor_id


async def test_remove_from_followers_following(sql_session_for_test):
    actor = await create_actor(sql_session_for_test, "http://localhost/ap/")

    remote_id = "http://remote.test"

    follower = Follower(actor=actor, follower=remote_id, accepted=True, request="")
    following = Following(actor=actor, following=remote_id, accepted=True, request="")

    sql_session_for_test.add_all([follower, following])
    await sql_session_for_test.commit()

    await remove_from_followers_following(sql_session_for_test, remote_id)

    assert 0 == await sql_session_for_test.scalar(func.count(Follower.id))
    assert 0 == await sql_session_for_test.scalar(func.count(Following.id))


async def test_identifier_ordering(sql_session_for_test):
    actor = await create_actor(sql_session_for_test, "http://localhost/ap/")

    one = PublicIdentifier(
        actor=actor, name="one", identifier="acct:jekyll@localhost", preference=1
    )
    two = PublicIdentifier(
        actor=actor, name="two", identifier="acct:hyde@localhost", preference=100
    )
    sql_session_for_test.add_all([one, two])
    await sql_session_for_test.commit()

    await sql_session_for_test.refresh(actor, attribute_names=["identifiers"])

    result = actor_to_object(actor)

    assert result["identifiers"] == [
        "acct:hyde@localhost",
        "acct:jekyll@localhost",
        actor.actor_id,
    ]
    assert result["preferredUsername"] == "hyde"

    one.preference = 1000
    await sql_session_for_test.merge(one)
    await sql_session_for_test.commit()
    await sql_session_for_test.refresh(actor, attribute_names=["identifiers"])

    result = actor_to_object(actor)

    assert result["identifiers"] == [
        "acct:jekyll@localhost",
        "acct:hyde@localhost",
        actor.actor_id,
    ]
    assert result["preferredUsername"] == "jekyll"


async def test_create_actor_duplicate_preferred_username(sql_session_for_test):
    await create_actor(
        sql_session_for_test, "http://localhost/ap/", preferred_username="me"
    )

    with pytest.raises(DuplicateIdentifierException):
        await create_actor(
            sql_session_for_test, "http://localhost/ap/", preferred_username="me"
        )


async def test_actor_to_object_attachments(sql_session_for_test):
    property_value_context = {
        "PropertyValue": {
            "@id": "https://schema.org/PropertyValue",
            "@context": {
                "value": "https://schema.org/value",
                "name": "https://schema.org/name",
            },
        }
    }
    property_value = {
        "type": "PropertyValue",
        "name": "key",
        "value": "value",
    }
    actor = await create_actor(
        sql_session_for_test,
        "http://localhost/ap/",
        profile={"attachment": [property_value]},
    )

    result = actor_to_object(actor)

    assert result["attachment"] == [property_value]
    assert property_value_context in result["@context"]



from uuid import UUID

from .models import PublishingActor
from .storage import object_for_name_and_uuid, publishing_actor_for_actor_id

from .testing import *  # noqa


async def test_publishing_actor_for_actor_id(sql_session_for_test):
    actor_id = "http://actor.test/some/id"

    result = await publishing_actor_for_actor_id(sql_session_for_test, actor_id)

    assert isinstance(result, PublishingActor)
    assert result.actor == actor_id


async def test_publishing_actor_for_actor_id_returns_stored_actor(sql_session_for_test):
    actor_id = "http://actor.test/some/id"

    one = await publishing_actor_for_actor_id(sql_session_for_test, actor_id)

    await sql_session_for_test.commit()

    two = await publishing_actor_for_actor_id(sql_session_for_test, actor_id)

    assert one.name == two.name


async def test_object_for_name_and_uuid_not_found(sql_session_for_test):
    result = await object_for_name_and_uuid(
        sql_session_for_test, "name", UUID("2fd16a00-309b-4f3a-9d91-aa9516e59c1f")
    )

    assert result is None



from cattle_grid.dependencies.globals import global_container

from cattle_grid.model.extension import (
    MethodInformationModel,
    AddMethodInformationMessage,
)


exchange_method_information = [
    MethodInformationModel(
        module="cattle_grid.exchange",
        routing_key="send_message",
        description="Takes an activity and sends it to its recipients",
    ),
    MethodInformationModel(
        module="cattle_grid.exchange",
        routing_key="update_actor",
        description="Updates an actor",
    ),
    MethodInformationModel(
        module="cattle_grid.exchange",
        routing_key="delete_actor",
        description="Deletes an actor",
    ),
]
"""Information about the methods defined on the ActivityExchange
by default"""


async def add_method_information(message: AddMethodInformationMessage):
    """Adds information about methods defined by an extension"""
    current_information = global_container.method_information or []

    global_container.method_information = (
        current_information + message.method_information
    )



import click

from cattle_grid.config.settings import get_settings
from . import add_to_cli


@click.group()
@click.pass_context
def main(ctx):
    ctx.ensure_object(dict)
    ctx.obj["config"] = get_settings()


add_to_cli(main)



import logging

from sqlalchemy import select, func
from sqlalchemy.ext.asyncio import AsyncSession
from faststream import Context
from faststream.rabbit import RabbitBroker

from bovine.activitystreams.utils import id_for_object

from cattle_grid.model import ActivityMessage as RawActivityMessage
from cattle_grid.database.activity_pub_actor import Actor, Blocking

from cattle_grid.dependencies import Transformer, CorrelationId, SqlSession
from cattle_grid.dependencies.globals import global_container

from cattle_grid.database.account import ActorForAccount

from cattle_grid.model.account import EventInformation, EventType

logger = logging.getLogger(__name__)


async def should_shovel_activity(session: AsyncSession, activity: dict) -> bool:
    """Some activities like Block or Undo Block should not be visible to the user. This method
    returns False if this is the case."""

    activity_type = activity.get("type")

    if activity_type == "Block":
        return False

    if activity_type == "Undo":
        object_id = id_for_object(activity.get("object"))
        blocking = await session.scalar(
            func.count(
                select(Blocking.id)
                .where(Blocking.request == object_id)
                .scalar_subquery()
            )
        )

        if blocking:
            return False

    return True


async def shovel_to_account_exchange(
    actor: str,
    account_name: str,
    event_type: EventType,
    to_shovel: dict,
    broker: RabbitBroker,
    correlation_id: str,
):
    await broker.publish(
        EventInformation(
            actor=actor,
            event_type=event_type,
            data=to_shovel,
        ),
        routing_key=f"receive.{account_name}.{event_type.value}",
        exchange=global_container.account_exchange,
        correlation_id=correlation_id,
    )


async def incoming_shovel(
    msg: RawActivityMessage,
    transformer: Transformer,
    correlation_id: CorrelationId,
    session: SqlSession,
    broker: RabbitBroker = Context(),
) -> None:
    """Transfers the message from the RawExchange to the
    Activity- and Account one.

    The message is passed through the transformer.
    """
    actor_for_account = await session.scalar(
        select(ActorForAccount).where(ActorForAccount.actor == msg.actor)
    )

    if actor_for_account is None:
        logger.warning("Got actor without account %s", msg.actor)
        return

    activity = msg.data

    if not await should_shovel_activity(session, activity):
        return

    # FIXME: Use join to combine the next two queries ...

    db_actor = await session.scalar(select(Actor).where(Actor.actor_id == msg.actor))
    if not db_actor:
        raise ValueError("Actor not found in database")

    blocking = await session.scalar(
        select(Blocking)
        .where(Blocking.actor == db_actor)
        .where(Blocking.blocking == activity.get("actor"))
        .where(Blocking.active)
    )
    if blocking:
        return

    account_name = actor_for_account.account.name
    to_shovel = await transformer({"raw": msg.data})
    activity_type = msg.data.get("type")

    await shovel_to_account_exchange(
        msg.actor, account_name, EventType.incoming, to_shovel, broker, correlation_id
    )

    await broker.publish(
        {
            "actor": msg.actor,
            "data": to_shovel,
        },
        routing_key=f"incoming.{activity_type}",
        exchange=global_container.exchange,
        correlation_id=correlation_id,
    )


async def outgoing_shovel(
    msg: RawActivityMessage,
    transformer: Transformer,
    correlation_id: CorrelationId,
    session: SqlSession,
    broker: RabbitBroker = Context(),
) -> None:
    actor_for_account = await session.scalar(
        select(ActorForAccount).where(ActorForAccount.actor == msg.actor)
    )

    logger.debug("OUTGOING SHOVEL")

    if actor_for_account is None:
        logger.warning("Actor %s for account not found", msg.actor)
        return

    account_name = actor_for_account.account.name
    to_shovel = await transformer({"raw": msg.data})
    activity_type = msg.data.get("type")

    await shovel_to_account_exchange(
        msg.actor, account_name, EventType.outgoing, to_shovel, broker, correlation_id
    )

    await broker.publish(
        {
            "actor": msg.actor,
            "data": to_shovel,
        },
        routing_key=f"outgoing.{activity_type}",
        exchange=global_container.exchange,
        correlation_id=correlation_id,
    )



from contextlib import asynccontextmanager
import logging

from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker

from .auth import Base as AuthBase


logger = logging.getLogger(__name__)

#
# Not sure the following function is what one wants
# ... it means pinning one database migrations work
# ... for
#


@asynccontextmanager
async def database_engine(
    db_uri: str = "sqlite+aiosqlite:///:memory:", echo: bool = False
):
    engine = create_async_engine(db_uri, echo=echo)

    yield engine

    await engine.dispose()


@asynccontextmanager
async def database_session(
    db_uri: str = "sqlite+aiosqlite:///:memory:", echo: bool = False
):
    async with database_engine(db_uri, echo) as engine:
        async_session = async_sessionmaker(engine)
        async with async_session() as session:
            yield session


async def upgrade_sql_alchemy(sql_engine):
    async with sql_engine.begin() as conn:
        await conn.run_sync(AuthBase.metadata.create_all)


async def run_with_database(config, coro):
    try:
        async with database_session(db_uri=config.db_uri) as session:
            await coro(session)
    except Exception as e:
        logger.exception(e)



from faststream import Context
from faststream.rabbit import RabbitBroker

from cattle_grid.account.account import group_names_for_actor
from cattle_grid.model.account import TriggerMessage
from cattle_grid.dependencies.globals import global_container
from cattle_grid.dependencies import CorrelationId, RewriteRules, SqlSession

from .annotations import ActorForAccountFromMessage, MethodFromRoutingKey


async def handle_trigger(
    msg: TriggerMessage,
    actor: ActorForAccountFromMessage,
    correlation_id: CorrelationId,
    method: MethodFromRoutingKey,
    session: SqlSession,
    rewrite_rules: RewriteRules,
    broker: RabbitBroker = Context(),
):
    """Used to trigger a method performed by the actor.

    The main thing an actor can do is send activities to
    the Fediverse. This can be done with `send_message`.
    This can be extended in cattle_grid through extensions.

    However the methods to update the actor profile and delete
    the actor are also called via a trigger.
    """

    if rewrite_rules:
        group_names = await group_names_for_actor(session, actor)
        method = rewrite_rules.rewrite(method, group_names)

    await broker.publish(
        msg,
        routing_key=method,
        exchange=global_container.exchange,
        correlation_id=correlation_id,
    )



import click
import asyncio
import logging

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.database import run_with_database
from cattle_grid.database.account import Account

# from cattle_grid.database import run_with_database
from .account import (
    create_account,
    add_permission,
    list_permissions,
    remove_permission,
    delete_account,
)

logger = logging.getLogger(__name__)


async def new_account(
    session: AsyncSession, name: str, password: str, permission: list[str]
):
    account = await create_account(session, name, password)
    if account is None:
        click.echo("Failed to create account")
        exit(1)

    for p in permission:
        await add_permission(session, account, p)


async def list_accounts(session: AsyncSession, actors):
    accounts = await Account.all().prefetch_related("permissions", "actors")
    for account in accounts:
        print(f"{account.name}: ", ", ".join(list_permissions(account)))
        if actors:
            for actor in account.actors:
                print(f"  {actor.name}: {actor.actor}")


async def modify_permissions(
    session: AsyncSession,
    name: str,
    add_permissions: list[str],
    remove_permissions: list[str],
):
    account = await session.scalar(select(Account).where(Account.name == name))
    if account is None:
        print(f"Account {name} does not exist")
        exit(1)
    for p in add_permissions:
        await add_permission(session, account, p)
    for p in remove_permissions:
        await remove_permission(session, account, p)


def add_account_commands(main):
    @main.group()
    def account():
        """Used to manage accounts associated with cattle_grid"""

    @account.command()  # type: ignore
    @click.argument("name")
    @click.argument("password")
    @click.option(
        "--admin", is_flag=True, default=False, help="Set the admin permission"
    )
    @click.option(
        "--permission",
        help="Adds the permission to the account",
        multiple=True,
        default=[],
    )
    @click.pass_context
    def new(ctx, name, password, admin, permission):
        """Creates a new account"""

        if admin:
            permission = list(permission) + ["admin"]

        asyncio.run(
            run_with_database(
                ctx.obj["config"],
                lambda session: new_account(session, name, password, permission),
            )
        )

    @account.command()  # type: ignore
    @click.argument("name")
    @click.option(
        "--add_permission",
        help="Adds the permission to the account",
        multiple=True,
        default=[],
    )
    @click.option(
        "--remove_permission",
        help="Adds the permission to the account",
        multiple=True,
        default=[],
    )
    @click.pass_context
    def modify(ctx, name, add_permission, remove_permission):
        """Modifies an account"""

        asyncio.run(
            run_with_database(
                ctx.obj["config"],
                lambda session: modify_permissions(
                    session,
                    name,
                    add_permissions=add_permission,
                    remove_permissions=remove_permission,
                ),
            )
        )

    @account.command("list")  # type: ignore
    @click.option(
        "--actors",
        is_flag=True,
        default=False,
        help="If set, also lists the actors associated with each account",
    )
    @click.pass_context
    def list_account(ctx, actors):
        """Lists existing accounts"""
        asyncio.run(
            run_with_database(
                ctx.obj["config"], lambda session: list_accounts(session, actors)
            )
        )

    @account.command("delete")  # type: ignore
    @click.argument("name")
    @click.pass_context
    def delete_account_command(ctx, name):
        """Lists existing accounts"""
        asyncio.run(
            run_with_database(
                ctx.obj["config"],
                lambda session: delete_account(session, name, force=True),
            )
        )



import logging

from faststream import Context
from faststream.rabbit import RabbitBroker


from cattle_grid.activity_pub.enqueuer import determine_activity_type
from cattle_grid.dependencies import CorrelationId
from cattle_grid.model import ActivityMessage, FetchMessage
from cattle_grid.dependencies.globals import global_container

logger = logging.getLogger(__name__)


async def send_message(
    msg: ActivityMessage,
    correlation_id: CorrelationId,
    broker: RabbitBroker = Context(),
) -> None:
    """Takes a message and ensure it is distributed appropriately

    FIXME: out_transformer?"""

    content = msg.data
    activity_type = determine_activity_type(content)

    if not activity_type:
        return

    to_send = ActivityMessage(actor=msg.actor, data=content)

    await broker.publish(
        to_send,
        exchange=global_container.exchange,
        routing_key=f"outgoing.{activity_type}",
        correlation_id=correlation_id,
    )
    await broker.publish(
        to_send,
        exchange=global_container.internal_exchange,
        routing_key=f"outgoing.{activity_type}",
        correlation_id=correlation_id,
    )


async def fetch_object(msg: FetchMessage, broker: RabbitBroker = Context()) -> dict:
    """Used to fetch an object as an RPC method"""
    result = await broker.publish(
        msg,
        routing_key="fetch_object",
        exchange=global_container.internal_exchange,
        rpc=True,
    )
    if result == b"" or result is None:
        return {}
    return result



import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.account.account import create_account

from .rabbit import rabbit_router


@pytest.fixture
def test_client():
    app = FastAPI()
    app.include_router(rabbit_router)

    yield TestClient(app)


async def test_user_unauthorized(test_client):
    response = test_client.post(
        "/rabbitmq/user", data={"username": "user", "password": "pass"}
    )

    assert response.status_code == 200
    assert response.text == "deny"


async def test_user(sql_session_for_test, test_client):
    await create_account(sql_session_for_test, "user", "pass")

    response = test_client.post(
        "/rabbitmq/user", data={"username": "user", "password": "pass"}
    )

    assert response.status_code == 200
    assert response.text == "allow"


@pytest.mark.parametrize(
    ["vhost", "result"],
    [("/", "allow"), ("/other", "deny"), ("/some/quite/long", "deny")],
)
async def test_vhost(test_client, vhost, result):
    response = test_client.post(
        "/rabbitmq/vhost", data={"username": "user", "vhost": vhost}
    )

    assert response.text == result


@pytest.mark.parametrize(
    ["name", "routing_key", "result"],
    [
        ("other", "send.user", "deny"),
        ("amq.topic", "other", "deny"),
        ("amq.topic", "send.user.trigger.send_message", "allow"),
        ("amq.topic", "receive.user.incoming", "allow"),
        ("amq.topic", "receive.user.#", "allow"),
        ("amq.topic", "receive.user_other.incoming", "deny"),
        ("amq.topic", "error.user", "allow"),
        ("amq.topic", "error.user_other", "deny"),
    ],
)
async def test_topic(test_client, name, routing_key, result):
    response = test_client.post(
        "/rabbitmq/topic",
        data={"username": "user", "name": name, "routing_key": routing_key},
    )

    assert response.text == result



from behave import then, when, given
from cattle_grid.testing.features import publish_as


@given('"{alice}" has the PropertyValue "{key}" with value "{value}"')
@when('"{alice}" adds the PropertyValue "{key}" with value "{value}"')
@when('"{alice}" updates the PropertyValue "{key}" with value "{value}"')
async def add_property_value(context, alice, key, value):
    await publish_as(
        context,
        alice,
        "update_actor",
        {
            "actor": context.actors[alice].get("id"),
            "actions": [
                {
                    "action": "update_property_value",
                    "key": key,
                    "value": value,
                }
            ],
        },
    )


@when('"{alice}" removes the PropertyValue "{key}"')
async def remove_property_value(context, alice, key):
    await publish_as(
        context,
        alice,
        "update_actor",
        {
            "actor": context.actors[alice].get("id"),
            "actions": [
                {
                    "action": "remove_property_value",
                    "key": key,
                }
            ],
        },
    )


@then('The profile contains the property value "{key}" with value "{value}"')
def check_property_value(context, key, value):
    attachments = context.result.get("attachment", [])
    assert isinstance(attachments, list)

    result = list(filter(lambda x: x["name"] == key, attachments))[0]

    assert result["type"] == "PropertyValue"
    assert result["value"] == value


@then('The profile does not contain the property value "{key}"')
def check_not_property_value(context, key):
    attachments = context.result.get("attachment", [])
    assert isinstance(attachments, list)

    filtered = list(filter(lambda x: x["name"] == key, attachments))

    assert len(filtered) == 0



from pydantic import ConfigDict, Field, field_serializer
from typing import Dict, Any, List

from .common import WithActor
from .exchange_update_actor import UpdateAction


class UpdateActorMessage(WithActor):
    """
    Allows one to update the actor object
    """

    # model_config = ConfigDict(
    #     extra="forbid",
    # )

    profile: Dict[str, Any] | None = Field(
        default=None,
        examples=[{"summary": "A new description of the actor"}],
        description="""
    New profile object for the actor. The fields.
    """,
    )
    autoFollow: bool | None = Field(
        default=None,
        examples=[True, False, None],
        description="""
    Enables setting actors to automatically accept follow requests
    """,
    )

    actions: List[UpdateAction] = Field(
        default_factory=list,
        description="""Actions to be taken when updating the profile""",
    )

    @field_serializer("actions")
    def serialize_dt(self, actions: List[UpdateAction], _info):
        return [action.model_dump() for action in actions]


class DeleteActorMessage(WithActor):
    """
    Allows one to delete the actor object
    """

    model_config = ConfigDict(
        extra="forbid",
    )



from .activity import actor_deletes_themselves


def test_actor_deletes_themselves():
    assert not actor_deletes_themselves({})


def test_actor_deletes_themselves_embedded_object():
    activity = {
        "type": "Delete",
        "actor": "http://actor.test/",
        "object": {
            "id": "http://actor.test/",
        },
    }

    assert actor_deletes_themselves(activity)



import pytest

from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.database.activity_pub_actor import (
    PublicIdentifier,
    PublicIdentifierStatus,
)

from .identifiers import collect_identifiers_for_actor, identifier_in_list_exists


def test_collect_identifiers_for_actor(actor_for_test):
    identifiers = collect_identifiers_for_actor(actor_for_test)

    assert identifiers == [actor_for_test.actor_id]


async def test_collect_identifiers_for_actor_with_acct_uri(
    sql_session_for_test, actor_for_test
):
    sql_session_for_test.add(
        PublicIdentifier(
            actor=actor_for_test,
            name="webfinger",
            identifier="acct:me@localhost",
            status=PublicIdentifierStatus.verified,
            preference=5,
        )
    )
    await sql_session_for_test.commit()
    await sql_session_for_test.refresh(actor_for_test, attribute_names=["identifiers"])

    identifiers = collect_identifiers_for_actor(actor_for_test)

    assert identifiers == ["acct:me@localhost", actor_for_test.actor_id]


async def test_collect_identifiers_for_actor_with_acct_uri_unverified(
    sql_session_for_test, actor_for_test
):
    sql_session_for_test.add(
        PublicIdentifier(
            actor=actor_for_test,
            name="webfinger",
            identifier="acct:me@localhost",
            status=PublicIdentifierStatus.unverified,
            preference=5,
        )
    )
    await sql_session_for_test.commit()
    await sql_session_for_test.refresh(actor_for_test, attribute_names=["identifiers"])

    identifiers = collect_identifiers_for_actor(actor_for_test)

    assert identifiers == [actor_for_test.actor_id]


@pytest.mark.parametrize(
    "identifiers, expected",
    [
        ([], False),
        (["acct:one@localhost"], True),
        (["acct:other@localhost"], False),
        (["acct:other@localhost", "acct:one@localhost"], True),
    ],
)
async def test_identifier_in_list_exists(
    session_maker_for_tests, actor_for_test, identifiers, expected
):
    async with session_maker_for_tests() as session:
        session.add(
            PublicIdentifier(
                actor_id=actor_for_test.id,
                name="webfinger",
                identifier="acct:one@localhost",
                status=PublicIdentifierStatus.unverified,
                preference=5,
            )
        )
        session.add(
            PublicIdentifier(
                actor_id=actor_for_test.id,
                name="webfinger",
                identifier="acct:two@localhost",
                status=PublicIdentifierStatus.unverified,
                preference=5,
            )
        )
        await session.commit()
    async with session_maker_for_tests() as session:
        result = await identifier_in_list_exists(session, identifiers)

    assert result == expected



from dataclasses import dataclass, field

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.account.account import (
    actor_for_actor_id,
    add_actor_to_group,
    group_names_for_actor,
)
from cattle_grid.database.account import ActorForAccount
from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.activity_pub.actor import actor_to_object


@dataclass
class ActorManager:
    """Access for managing actors from outside cattle_grid, e.g.
    by an extension"""

    session: AsyncSession
    actor_id: str = field(metadata={"description": "The URI of the actor to manage"})

    _actor_for_account: ActorForAccount | None = field(default=None)
    _actor: Actor | None = field(default=None)

    async def actor_for_account(self) -> ActorForAccount:
        if not self._actor_for_account:
            self._actor_for_account = await actor_for_actor_id(
                self.session, self.actor_id
            )

        if not self._actor_for_account:
            raise ValueError("Actor not found")

        return self._actor_for_account

    async def actor(self, session: AsyncSession) -> Actor:
        if not self._actor:
            self._actor = await session.scalar(
                select(Actor).where(Actor.actor_id == self.actor_id)
            )
            await session.refresh(self._actor, attribute_names=["identifiers"])

        if not self._actor:
            raise ValueError("Actor not found")
        return self._actor

    async def add_to_group(self, group: str):
        """Adds the actor to a group"""
        actor = await self.actor_for_account()
        await add_actor_to_group(self.session, actor, group)

    async def groups(self) -> list[str]:
        """Returns the list of groups the actor belongs to"""
        actor = await self.actor_for_account()
        return await group_names_for_actor(self.session, actor)

    async def profile(self, session: AsyncSession) -> dict:
        """Returns the actor profile"""
        return actor_to_object(await self.actor(session))



import pytest

from .format import format_actor_profile


@pytest.fixture
def profile():
    return {
        "attachment": [
            {
                "type": "PropertyValue",
                "name": "Author",
                "value": "acct:helge@mymath.rocks",
            },
            {
                "type": "PropertyValue",
                "name": "Source",
                "value": "https://codeberg.org/bovine/roboherd",
            },
            {
                "type": "PropertyValue",
                "name": "Frequency",
                "value": "At 42 minutes past the hour",
            },
        ],
        "published": "2025-09-07T12:39:49.096182",
        "@context": [
            "https://www.w3.org/ns/activitystreams",
            "https://w3id.org/security/v1",
            {
                "PropertyValue": {
                    "@id": "https://schema.org/PropertyValue",
                    "@context": {
                        "value": "https://schema.org/value",
                        "name": "https://schema.org/name",
                    },
                }
            },
        ],
        "publicKey": {
            "id": "http://abel/actor/FqHFi7vFmVOalVXogLZmtA#legacy-key-1",
            "owner": "http://abel/actor/FqHFi7vFmVOalVXogLZmtA",
            "publicKeyPem": "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA4wMPZRHfpDSTL2p0Y/Fo\nZ7VCUYh+f0NUqobOQgwEacdetNwcmgIwYwldu5mpbM5vKlTEac9fYSK6r9yiIj2q\nsds9CvyHXfTeiapI4nYWOtnLzwbW1p4t0FOpcwii6s4flii937CYGVgCmNyB1UR9\nqDbdbJwVUCeO9oixOnt2WWO5EZnBWGpMnCfNtbGiRdFQthCWia+KWMkcle6xSxWk\n+69Xiu4ClH0ma9TfftC4eEZgZ8UQdmLgdOKKysGk+5+jJBClqQwZOjhRWBZtH+Js\nPmV33KKrK5nd0sDqekExXqsJ23UDTVcXywhIBtFy948VSVtlmBM9jX9wzZJ/i8Tf\nqQIDAQAB\n-----END PUBLIC KEY-----\n",
        },
        "id": "http://abel/actor/FqHFi7vFmVOalVXogLZmtA",
        "type": "Service",
        "inbox": "http://abel/inbox/8JF7O5ADRHFaRoHSKVRc9Q",
        "outbox": "http://abel/outbox/iWm7WPxhdWJOC-9CIWArlw",
        "followers": "http://abel/followers/zUK1yGIj0KDI611BK7859g",
        "following": "http://abel/following/Lu7i17EUCg7zWjNjBt-z1A",
        "preferredUsername": "rooster",
        "name": "The crowing rooster \ud83d\udc13",
        "icon": {
            "mediaType": "image/png",
            "type": "Image",
            "url": "https://dev.bovine.social/assets/bull-horns.png",
        },
        "identifiers": [
            "acct:rooster@abel",
            "http://abel/actor/FqHFi7vFmVOalVXogLZmtA",
        ],
        "endpoints": {"sharedInbox": "http://abel/shared_inbox"},
        "summary": "I'm an actor",
    }


def test_format_actor_profile_empty_input():
    format_actor_profile({})


def test_format_actor_profile_fediverse_handles(profile):
    formatted = format_actor_profile(profile)

    assert formatted["fediverse_handles"] == ["@rooster@abel"]


def test_format_actor_profile_summary(profile):
    formatted = format_actor_profile(profile)

    assert formatted["summary"] == "I'm an actor"


def test_format_actor_profile_property_values(profile):
    formatted = format_actor_profile(profile)

    assert formatted["property_values"] == [
        {
            "name": "Author",
            "value": "acct:helge@mymath.rocks",
        },
        {
            "name": "Source",
            "value": "https://codeberg.org/bovine/roboherd",
        },
        {
            "name": "Frequency",
            "value": "At 42 minutes past the hour",
        },
    ]



from pydantic import BaseModel, Field


class MethodInformationModel(BaseModel):
    """cattle_grid allows to define methods on the
    exchange through extensions. This class contains
    a description of them"""

    routing_key: str = Field(
        examples=["send_message"],
        description="""Name of the method""",
    )

    module: str = Field(
        examples=["cattle_grid"],
        description="""Module the extension was imported from. This is cattle_grid for build-in methods""",
    )

    description: str | None = Field(
        default=None,
        examples=["Send a message as the actor"],
        description="""Description of the method""",
    )


class AddMethodInformationMessage(BaseModel):
    """Message to add method information through the exchange"""

    method_information: list[MethodInformationModel] = Field(
        description="""List of method information""",
    )



import asyncio
from behave import when, then

from bovine.activitystreams import factories_for_actor_object

from cattle_grid.testing.features import publish_as, fetch_request


@when('"{alice}" publishes a "{moo}" animal sound to her followers')  # pyright: ignore[reportCallIssue]
async def send_sound(context, alice, moo):
    for connection in context.connections.values():
        await connection.clear_incoming()

    alice_actor = context.actors[alice]
    activity_factory, _ = factories_for_actor_object(alice_actor)

    activity = (
        activity_factory.custom(type="AnimalSound", content="moo").as_public().build()
    )

    await publish_as(
        context,
        alice,
        "publish_activity",
        {"actor": context.actors[alice].get("id"), "data": activity},
    )


@when('"{alice}" publishes a message "{text}" to her followers')  # pyright: ignore[reportCallIssue]
async def send_message(context, alice, text):
    await asyncio.sleep(0.3)
    for connection in context.connections.values():
        await connection.clear_incoming()

    alice_actor = context.actors[alice]
    _, object_factory = factories_for_actor_object(alice_actor)

    note = object_factory.note(content=text).as_public().build()

    await publish_as(
        context,
        alice,
        "publish_object",
        {"actor": context.actors[alice].get("id"), "data": note},
    )


@then('"{bob}" can retrieve the activity')  # pyright: ignore[reportCallIssue]
async def can_retrieve_activity(context, bob):
    result = await fetch_request(context, bob, context.activity.get("id"))
    assert result



import asyncio
from behave import given, then

from cattle_grid.extensions.examples.html_display.types import NameActorMessage


@given('"{alice}" sets her display name to "{alex}"')  # type: ignore
async def html_set_display_name(context, alice, alex):
    connection = context.connections[alice]
    alice_id = context.actors[alice].get("id")

    await connection.trigger(
        "html_display_name", NameActorMessage(actor=alice_id, name=alex).model_dump()
    )

    await asyncio.sleep(0.2)


@then("The profile contains an url")  # type: ignore
def profile_contains_url(context):
    urls = context.profile.get("url")

    print(urls)
    print(context.profile)

    assert isinstance(urls, list)
    assert len(urls) >= 1



from dataclasses import dataclass
from typing import Callable, Literal
import logging


from bovine import BovineActor
from bovine.crypto.types import CryptographicIdentifier
from fediverse_pasture.server.utils import actor_object_to_public_key
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select

from cattle_grid.database.auth import RemoteIdentity

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def result_for_identity(identity: RemoteIdentity):
    if identity.controller == "gone":
        return None

    return CryptographicIdentifier.from_tuple(identity.controller, identity.public_key)


@dataclass
class PublicKeyCache:
    """Caches public keys in the database and fetches them
    using bovine_actor"""

    bovine_actor: BovineActor
    """used to fetch the public key"""

    session_maker: Callable[[], AsyncSession] | None = None
    """sql session maker"""

    async def cryptographic_identifier(
        self, key_id: str
    ) -> CryptographicIdentifier | Literal["gone"] | None:
        """Returns "gone" if Tombstone

        :param key_id: URI of the public key to fetch
        :returns:
        """

        try:
            result = await self.bovine_actor.get(key_id)

            if result is None:
                return "gone"

            if result.get("type") == "Tombstone":
                logger.info("Got Tombstone for %s", key_id)
                return "gone"

            public_key, owner = actor_object_to_public_key(result, key_id)

            if public_key is None or owner is None:
                return None

            return CryptographicIdentifier.from_pem(public_key, owner)
        except Exception as e:
            logger.info("Failed to fetch public key for %s with %s", key_id, repr(e))
            logger.exception(e)
            return None

    async def from_cache(self, key_id: str) -> CryptographicIdentifier | None:
        if not self.session_maker:
            raise Exception("Sql session must be set")

        async with self.session_maker() as session:
            try:
                identity = await session.scalar(
                    select(RemoteIdentity).where(RemoteIdentity.key_id == key_id)
                )
            except Exception as e:
                logger.exception(e)
                identity = None

            if identity:
                return result_for_identity(identity)

            identifier = await self.cryptographic_identifier(key_id)
            if identifier is None:
                return None

            if identifier == "gone":
                session.add(
                    RemoteIdentity(key_id=key_id, public_key="gone", controller="gone")
                )
                await session.commit()
                return None

            try:
                controller, public_key = identifier.as_tuple()

                session.add(
                    RemoteIdentity(
                        key_id=key_id, public_key=public_key, controller=controller
                    )
                )
                await session.commit()
            except Exception as e:
                logger.exception(e)
            return identifier



import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from sqlalchemy.orm.attributes import flag_modified

from .router import ap_router

from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.activity_pub.actor import delete_actor
from cattle_grid.database.activity_pub_actor import Follower, Blocking
from cattle_grid.activity_pub.test_enqueuer import mock_broker  # noqa


default_requester = {"headers": {"x-cattle-grid-requester": "remote"}}


@pytest.fixture
def test_client():
    app = FastAPI()
    app.include_router(ap_router)

    yield TestClient(app)


def endpoint_from_actor(actor, endpoint):
    if endpoint == "actor":
        return actor.actor_id
    else:
        return actor.__getattribute__(f"{endpoint}_uri")


@pytest.mark.parametrize("endpoint", ["actor", "outbox", "followers", "following"])
async def test_successful_get(test_client, actor_for_test, endpoint):
    path = endpoint_from_actor(actor_for_test, endpoint)

    response = test_client.get(
        path,
        headers={
            "x-cattle-grid-requester": "owner",
            "x-ap-location": path,
        },
    )

    assert response.status_code == 200


@pytest.mark.parametrize("endpoint", ["actor", "outbox", "followers", "following"])
async def test_deleted_actor(
    sql_session_for_test, test_client, actor_for_test, endpoint
):  # noqa
    path = endpoint_from_actor(actor_for_test, endpoint)
    await delete_actor(sql_session_for_test, actor_for_test)

    response = test_client.get(
        path,
        headers={
            "x-cattle-grid-requester": "owner",
            "x-ap-location": path,
        },
    )

    assert response.status_code == 410


@pytest.mark.parametrize("endpoint", ["actor", "outbox", "followers", "following"])
async def test_endpoint_not_found(sql_session_for_test, test_client, endpoint):
    response = test_client.get(
        f"http://localhost/{endpoint}/not_an_actor",
        headers={
            "x-cattle-grid-requester": "owner",
            "x-ap-location": f"http://localhost/{endpoint}/not_an_actor",
        },
    )

    assert response.status_code == 404


@pytest.mark.parametrize("endpoint", ["actor", "outbox", "followers", "following"])
async def test_retriever_is_blocked(
    sql_session_for_test, test_client, actor_for_test, endpoint
):  # noqa
    path = endpoint_from_actor(actor_for_test, endpoint)
    sql_session_for_test.add(
        Blocking(actor=actor_for_test, blocking="owner", request="some_id", active=True)
    )

    await sql_session_for_test.commit()

    response = test_client.get(
        path,
        headers={
            "x-cattle-grid-requester": "owner",
            "x-ap-location": path,
        },
    )

    assert response.status_code == 403


@pytest.mark.parametrize("endpoint", ["actor", "outbox", "followers", "following"])
async def test_retriever_is_blocked_inactive(
    sql_session_for_test, test_client, actor_for_test, endpoint
):  # noqa
    path = endpoint_from_actor(actor_for_test, endpoint)
    sql_session_for_test.add(
        Blocking(
            actor=actor_for_test, blocking="owner", request="some_id", active=False
        )
    )
    await sql_session_for_test.commit()

    response = test_client.get(
        path,
        headers={
            "x-cattle-grid-requester": "owner",
            "x-ap-location": path,
        },
    )

    assert response.status_code == 200


@pytest.mark.parametrize("endpoint", ["outbox", "followers", "following"])
async def test_collection(test_client, endpoint, actor_for_test):  # noqa
    path = actor_for_test.__getattribute__(f"{endpoint}_uri")

    response = test_client.get(
        path,
        headers={
            "x-cattle-grid-requester": "owner",
            "x-ap-location": path,
        },
    )

    assert response.status_code == 200
    assert response.headers["content-type"] == "application/activity+json"

    data = response.json()

    assert data["type"] == "OrderedCollection"
    assert data["id"] == path


async def test_actor(test_client, actor_for_test):  # noqa
    response = test_client.get(
        actor_for_test.actor_id,
        headers={
            "x-cattle-grid-requester": "owner",
            "x-ap-location": actor_for_test.actor_id,
        },
    )

    assert response.status_code == 200
    data = response.json()

    assert data["id"] == actor_for_test.actor_id

    assert data["identifiers"] == [actor_for_test.actor_id]


@pytest.mark.parametrize("endpoint", ["actor", "outbox", "followers", "following"])
async def test_unauthorized(endpoint, test_client, actor_for_test):  # noqa
    path = endpoint_from_actor(actor_for_test, endpoint)
    response = test_client.get(
        path,
        headers={
            "x-ap-location": path,
        },
    )

    assert response.status_code == 401


async def test_followers(sql_session_for_test, test_client, actor_for_test):  # noqa
    async def follower_info():
        response = test_client.get(
            actor_for_test.followers_uri,
            headers={
                "x-cattle-grid-requester": "owner",
                "x-ap-location": actor_for_test.followers_uri,
            },
        )

        assert response.status_code == 200

        return response.json()

    async def follower_count():
        data = await follower_info()
        return data.get("totalItems", 0)

    assert 0 == await follower_count()

    follower = Follower(
        actor=actor_for_test,
        follower="http://follower",
        request="http://follower/id",
        accepted=False,
    )

    sql_session_for_test.add(follower)
    await sql_session_for_test.commit()

    assert 0 == await follower_count()

    follower.accepted = True
    await sql_session_for_test.merge(follower)
    await sql_session_for_test.commit()

    assert 1 == await follower_count()

    data = await follower_info()

    assert data["orderedItems"] == ["http://follower"]


async def test_actor_html_request(test_client, actor_for_test):
    response = test_client.get(
        actor_for_test.actor_id,
        headers={
            "x-cattle-grid-requester": "owner",
            "x-ap-location": actor_for_test.actor_id,
            "x-cattle-grid-should-serve": "html",
        },
    )

    assert response.status_code == 406


async def test_actor_html_request_has_url(
    test_client, sql_session_for_test, actor_for_test
):
    actor_for_test.profile["url"] = [
        {
            "type": "Link",
            "href": "http://html-server.test/actor",
            "mediaType": "text/html",
        }
    ]
    flag_modified(actor_for_test, "profile")
    await sql_session_for_test.commit()

    response = test_client.get(
        actor_for_test.actor_id,
        headers={
            "x-cattle-grid-requester": "owner",
            "x-ap-location": actor_for_test.actor_id,
            "x-cattle-grid-should-serve": "html",
        },
        follow_redirects=False,
    )

    assert response.status_code == 307



from urllib.request import urlopen
from urllib.parse import urlparse
from typing import Set
import logging

from bovine import BovineActor
from cattle_grid.config.auth import AuthConfig

logger = logging.getLogger(__name__)


def blocklist_form_url_or_file(url_or_file: str) -> Set[str]:
    if url_or_file.startswith("https://"):
        with urlopen(url_or_file) as f:
            data = [x.decode("utf-8") for x in f.readlines()]
    else:
        with open(url_or_file) as f:
            data = f.readlines()

    result = {
        x.removesuffix("\n") for x in data if x != "canary.fedinuke.example.com\n"
    }

    return result


def config_to_bovine_actor(auth_config: AuthConfig) -> BovineActor:
    return BovineActor(
        actor_id=auth_config.actor_id,
        secret=auth_config.private_key,
        public_key_url=auth_config.actor_id + "#mykey",
    )


def check_block(domain_blocks: Set[str], controller: str) -> bool:
    """Checks if a controller's domain is in block list

    ```pycon
    >>> check_block({"blocked.example"}, "http://actor.example/path")
    False

    >>> check_block({"blocked.example"}, "http://blocked.example/path")
    True

    ```
    """
    try:
        domain = urlparse(controller).netloc
        return domain in domain_blocks
    except Exception as e:
        logger.warning("Something went wrong with %s", repr(e))
        return True



from faststream.rabbit import RabbitRouter, RabbitExchange, RabbitQueue
from typing import Any, List, Tuple

from cattle_grid.dependencies.globals import global_container

from .remote import fetch_object, sending_message
from .outgoing import create_outgoing_router
from .incoming import create_incoming_router
from .store_activity import store_activity_subscriber
from .shared_inbox import handle_shared_inbox_message


def create_processing_router(exchange: RabbitExchange | None = None) -> RabbitRouter:
    if exchange is None:
        exchange = global_container.internal_exchange

    router = RabbitRouter()
    router.include_router(create_outgoing_router(exchange))
    router.include_router(create_incoming_router(exchange))

    routing_config: List[Tuple[str, Any]] = [
        ("store_activity", store_activity_subscriber),
        ("to_send", sending_message),
        ("fetch_object", fetch_object),
        ("shared_inbox", handle_shared_inbox_message),
    ]
    for routing_key, coroutine in routing_config:
        router.subscriber(
            RabbitQueue(f"cg_internal_{routing_key}", routing_key=routing_key),
            exchange=exchange,
            title=f"Internal:{routing_key}",
        )(coroutine)

    return router



from . import load_extension

from .transformer import resolvable_transformers, get_transformers
from .types import Transformer


async def transformer(d: dict) -> dict:
    return d


one = Transformer(
    name="one",
    transformer=transformer,
    inputs={"raw"},
    outputs={"one"},
)

two = Transformer(
    name="two",
    transformer=transformer,
    inputs={"one"},
    outputs={"two"},
)


def test_resolvable_transformer_one():
    assert resolvable_transformers({one}, {"raw"}) == {one}


def test_get_transformers():
    extensions = load_extension(
        {"module": "cattle_grid.extensions.examples.recipients"}
    )

    transformers = get_transformers([extensions])

    assert len(transformers) == 1
    assert isinstance(list(transformers)[0], Transformer)


# FIXME: Tests for dependency injection



import datetime

from sqlalchemy import String, func, JSON
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from sqlalchemy_utils.types import UUIDType


class Base(AsyncAttrs, DeclarativeBase):
    """Base model"""

    pass


class PublishingActor(Base):
    __tablename__ = "html_display_publishing_actor"

    id: Mapped[int] = mapped_column(primary_key=True)

    actor: Mapped[str] = mapped_column(String(256))
    name: Mapped[str] = mapped_column(String(256), unique=True)


class PublishedObject(Base):
    """HTML display object in the database"""

    __tablename__ = "html_display_stored_object"
    """name of the table"""

    id: Mapped[bytes] = mapped_column(UUIDType(binary=True), primary_key=True)
    """The id (uuid as bytes)"""
    data: Mapped[dict] = mapped_column(JSON)
    """The object as JSON"""
    actor: Mapped[str] = mapped_column()
    """The actor that created the object"""
    create_date: Mapped[datetime.datetime] = mapped_column(server_default=func.now())
    """The create timestamp"""



from unittest.mock import AsyncMock

from sqlalchemy import select

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.model.account import TriggerMessage

from cattle_grid.account.account import add_actor_to_group
from cattle_grid.database.account import ActorForAccount
from cattle_grid.config.rewrite import RewriteConfiguration

from .trigger import handle_trigger


async def test_handle_trigger(sql_session_for_test, actor_with_account):
    broker = AsyncMock()
    actor_for_account = await sql_session_for_test.scalar(
        select(ActorForAccount).where(
            ActorForAccount.actor == actor_with_account.actor_id
        )
    )
    assert actor_for_account

    await handle_trigger(
        TriggerMessage(
            actor=actor_with_account.actor_id,  # type:ignore
        ),
        session=sql_session_for_test,
        actor=actor_for_account,
        broker=broker,
        correlation_id="uuid",
        method="method",
        rewrite_rules=RewriteConfiguration.from_rules({}),
    )

    broker.publish.assert_awaited_once()

    (_, kwargs) = broker.publish.call_args

    assert kwargs["routing_key"] == "method"
    assert kwargs["correlation_id"] == "uuid"


async def test_handle_trigger_with_rewrite(sql_session_for_test, actor_with_account):
    broker = AsyncMock()
    actor_for_account = await sql_session_for_test.scalar(
        select(ActorForAccount).where(
            ActorForAccount.actor == actor_with_account.actor_id
        )
    )
    assert actor_for_account

    await add_actor_to_group(sql_session_for_test, actor_for_account, "group")

    rewrite_config = {
        "group": {
            "method": "changed",
        }
    }

    await handle_trigger(
        TriggerMessage(
            actor=actor_with_account.actor_id,
        ),
        session=sql_session_for_test,
        actor=actor_for_account,
        broker=broker,
        correlation_id="uuid",
        method="method",
        rewrite_rules=RewriteConfiguration.from_rules(rewrite_config),
    )

    broker.publish.assert_awaited_once()

    (_, kwargs) = broker.publish.call_args

    assert kwargs["routing_key"] == "changed"
    assert kwargs["correlation_id"] == "uuid"



from cattle_grid.model.extension import (
    MethodInformationModel,
    AddMethodInformationMessage,
)

from .info import add_method_information


async def test_add_method_information():
    message = AddMethodInformationMessage(
        method_information=[
            MethodInformationModel(
                module="testing",
                routing_key="test_key",
            ),
        ]
    )

    await add_method_information(message)



from sqlalchemy import func
from .database import database_session
from .database.auth import RemoteIdentity
from cattle_grid.database.activity_pub_actor import Actor


async def statistics(config):
    async with database_session(db_uri=config.db_uri) as session:
        remote_identity_count = await session.scalar(func.count(RemoteIdentity.id))
        actor_count = await session.scalar(func.count(Actor.id))

        print(f"Remote identity count: {remote_identity_count:10d}")
        print(f"Actor count:           {actor_count:10d}")



import logging

from typing import Annotated
from faststream import Context
from fast_depends import Depends

from bovine.activitystreams import factories_for_actor_object
from bovine.activitystreams.activity_factory import ActivityFactory
from bovine.activitystreams.object_factory import ObjectFactory
from sqlalchemy import select


from cattle_grid.activity_pub.actor import actor_to_object
from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.dependencies import SqlSession
from cattle_grid.model.common import WithActor


logger = logging.getLogger(__name__)


RoutingKey = Annotated[str, Context("message.raw_message.routing_key")]
"""The AMQP routing key"""


class ProcessingError(ValueError): ...


async def actor_id(message: WithActor) -> str:
    return message.actor


async def actor_for_message(session: SqlSession, actor_id: str = Depends(actor_id)):
    actor = await session.scalar(select(Actor).where(Actor.actor_id == actor_id))

    if actor is None:
        raise ProcessingError("Actor not found")

    return actor


MessageActor = Annotated[Actor, Depends(actor_for_message)]
"""Returns the actor for the message"""


def get_actor_profile(actor: MessageActor):
    return actor_to_object(actor)


ActorProfile = Annotated[dict, Depends(get_actor_profile)]
"""Returns the actor profile of the actor processing the
message"""


def get_factories_for_actor(profile: ActorProfile):
    return factories_for_actor_object(profile)


FactoriesForActor = Annotated[
    tuple[ActivityFactory, ObjectFactory], Depends(get_factories_for_actor)
]
"""Returns the activity and object factories for the actor"""



import logging

from faststream import Context
from faststream.rabbit import RabbitRouter
from sqlalchemy import select, delete, update

from bovine.activitystreams.utils import (
    recipients_for_object,
    id_for_object,
    remove_public,
)

from cattle_grid.database.activity_pub_actor import Following, Follower, Blocking
from cattle_grid.model import ActivityMessage
from cattle_grid.model.processing import ToSendMessage
from cattle_grid.dependencies import CommittingSession, SqlSession
from cattle_grid.dependencies.globals import global_container
from cattle_grid.dependencies.processing import MessageActor

from .util import update_recipients_for_collections

logger = logging.getLogger(__name__)


async def outgoing_message_distribution(
    message: ActivityMessage, session: SqlSession, broker=Context()
):
    """Distributes the message to its recipients"""

    recipients = recipients_for_object(message.data)
    recipients = remove_public(recipients)

    logger.debug("Got recipients %s", ", ".join(recipients))

    recipients = await update_recipients_for_collections(session, message, recipients)

    for recipient in recipients:
        await broker.publish(
            ToSendMessage(actor=message.actor, data=message.data, target=recipient),
            exchange=global_container.internal_exchange,
            routing_key="to_send",
        )


async def outgoing_follow_request(
    message: ActivityMessage,
    actor: MessageActor,
    session: CommittingSession,
    broker=Context(),
):
    """Handles an outgoing Follow request"""

    follow_request = message.data
    to_follow = follow_request.get("object")
    if isinstance(to_follow, dict):
        to_follow = to_follow.get("id")

    if to_follow is None:
        return

    logger.info("Send follow request to %s", to_follow)

    session.add(
        Following(
            actor=actor,
            following=to_follow,
            request=follow_request.get("id"),
            accepted=False,
        )
    )


async def outgoing_accept_request(
    message: ActivityMessage,
    actor: MessageActor,
    session: CommittingSession,
    broker=Context(),
):
    """Handles an outgoing Accept activity"""
    accept_request = message.data
    request_being_accepted = id_for_object(accept_request.get("object"))

    follower = await session.scalar(
        select(Follower).where(Follower.request == request_being_accepted)
    )

    if not follower:
        logger.warning("Follow request with id '%s' not found", request_being_accepted)
        return

    follower.accepted = True
    logger.info("Accepted follow request %s", request_being_accepted)


async def outgoing_undo_request(
    message: ActivityMessage,
    actor: MessageActor,
    session: CommittingSession,
    broker=Context(),
):
    """Handles an outgoing Undo activity"""
    accept_request = message.data
    request_being_undone = accept_request.get("object")

    await session.execute(
        delete(Following).where(Following.request == request_being_undone)
    )
    await session.execute(
        update(Blocking)
        .where(Blocking.request == request_being_undone)
        .values(active=False)
    )


async def outgoing_reject_activity(
    message: ActivityMessage,
    actor: MessageActor,
    session: CommittingSession,
    broker=Context(),
):
    """Handles an outgoing Reject activity"""
    reject_request = message.data
    request_being_rejected = reject_request.get("object")

    await session.execute(
        delete(Follower).where(Follower.request == request_being_rejected)
    )


async def outgoing_block_activity(
    message: ActivityMessage,
    actor: MessageActor,
    session: CommittingSession,
    broker=Context(),
):
    """Handles an outgoing Block activity"""
    block_request = message.data
    actor_being_blocked = block_request.get("object")

    await session.execute(
        delete(Follower).where(Follower.follower == actor_being_blocked)
    )

    block_id = block_request.get("id", "permanent")

    if block_id == "permanent":
        logger.warning("%s permanently blocked %s", actor.actor_id, actor_being_blocked)

    session.add(
        Blocking(
            actor=actor, blocking=actor_being_blocked, request=block_id, active=True
        )
    )
    logger.info("%s blocked %s", actor.actor_id, actor_being_blocked)


def create_outgoing_router(exchange=None):
    router = RabbitRouter()

    if exchange is None:
        exchange = global_container.internal_exchange

    for routing_key, coroutine in [
        ("outgoing.Follow", outgoing_follow_request),
        ("outgoing.Accept", outgoing_accept_request),
        ("outgoing.Undo", outgoing_undo_request),
        ("outgoing.Block", outgoing_block_activity),
        ("outgoing.Reject", outgoing_reject_activity),
        ("outgoing.#", outgoing_message_distribution),
    ]:
        router.subscriber(routing_key, exchange=exchange, title=routing_key)(coroutine)

    return router



import pytest
from fastapi.testclient import TestClient
from unittest.mock import patch
from sqlalchemy.ext.asyncio import async_sessionmaker

from bovine.crypto.types import CryptographicIdentifier
from bovine.testing import public_key

from cattle_grid.testing.fixtures import auth_config  # noqa
from . import create_app
from .test_public_key_cache import *  # noqa

from cattle_grid.database.auth import RemoteIdentity, Base
from .dependencies import provide_auth_config


def valid_headers_for_key_id(key_id):
    return {
        "date": "Wed, 15 Mar 2023 17:28:15 GMT",
        "X-Original-Host": "myhost.example",
        "X-Original-Uri": "/path/to/resource",
        "X-Original-Method": "get",
        "signature": f'''keyId="{key_id}",algorithm="rsa-sha256",headers="(request-target) host date",signature="hUW2jMUkhiKTmAoqgq7CDz0l4nYiulbVNZflKu0Rxs34FyBs0zkBKLZLUnR35ptOvsZA7hyFOZbmK9VTw2VnoCvUYDPUb5VyO3MRpLv0pfXNExQEWuBMEcdvXTo30A0WIDSL95u7a6sQREjKKHD5+edW85WhhkqhPMtGpHe95cMItIBv6K5gACrsOYf8TyhtYqBxz8Et0iwoHnMzMCAHN4C+0nsGjqIfxlSqUSMrptjjov3EBEnVii9SEaWCH8AUE9kfh3FeZkT+v9eIDZdhj4+opnJlb9q2+7m/7YH0lxaXmqro0fhRFTd832wY/81LULix/pWTOmuJthpUF9w6jw=="''',
    }


@pytest.fixture
def test_client(auth_config, sql_engine_for_tests):  # noqa
    app = create_app()

    app.dependency_overrides[provide_auth_config] = lambda: auth_config

    yield TestClient(app)


@pytest.fixture
async def test_session(sql_engine_for_tests):
    async with sql_engine_for_tests.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)

    session_maker = async_sessionmaker(sql_engine_for_tests)

    async with session_maker() as session:
        yield session


# async def test_get_index(test_client):
#     response = await test_client.get("/")

#     assert response.status_code == 200

#     text = await response.get_data()

#     assert "cattle_grid" in text.decode("utf-8")


async def test_get_actor(test_client):
    response = test_client.get("/cattle_grid_actor")

    assert response.status_code == 200

    result = response.json()

    assert response.headers["content-type"] == "application/activity+json"
    assert result["preferredUsername"] == "actor"


async def test_webfinger(test_client):
    response = test_client.get("/.well-known/webfinger?resource=acct:actor@localhost")

    assert response.status_code == 200

    result = response.json()
    print(result)

    assert result["subject"] == "acct:actor@localhost"


async def test_webfinger_not_found(test_client):
    response = test_client.get("/.well-known/webfinger?resource=acct:other@localhost")

    assert response.status_code == 404


async def test_webfinger_bad_request(test_client):
    response = test_client.get("/.well-known/webfinger")

    assert response.status_code == 422


@patch("bovine.utils.check_max_offset_now", lambda x: True)
async def test_get_auth(test_client):
    response = test_client.get("/auth")

    assert response.status_code == 200


async def test_get_auth_invalid_signature(test_client):
    response = test_client.get("/auth", headers={"Signature": "invalid"})

    assert response.status_code == 401


@patch("bovine.utils.check_max_offset_now", lambda x: True)
async def test_get_auth_invalid_signature_cannot_get_key(test_client):
    response = test_client.get(
        "/auth",
        headers={
            "Signature": '''keyId="https://remote.example/actor#key",algorithm="rsa-sha256",headers="(request-target) host",signature="inv sfdsfalid=="''',
            "X-Original-Method": "GET",
            "X-Original-Host": "remote.example",
            "X-Original-Uri": "/path",
            "Date": "today",
        },
    )

    assert response.status_code == 401


@patch("bovine.utils.check_max_offset_now", lambda x: True)
async def test_get_auth_invalid_signature_can_get_key(test_client, test_session):
    controller = "https://remote.example/actor"
    key_id = f"{controller}#key"
    identifier = CryptographicIdentifier.from_pem(public_key, controller)
    public_key_multibase = identifier.as_tuple()[1]
    test_session.add(
        RemoteIdentity(
            key_id=key_id,
            controller=controller,
            public_key=public_key_multibase,
        )
    )
    await test_session.commit()

    response = test_client.get(
        "/auth",
        headers=valid_headers_for_key_id(key_id),
    )

    assert response.status_code == 200

    assert response.headers["X-Cattle-Grid-Requester"] == controller


@patch("bovine.utils.check_max_offset_now", lambda x: True)
async def test_get_auth_invalid_signature_can_get_key_blocked(
    test_client, test_session
):
    controller = "https://blocked.example/actor"
    key_id = f"{controller}#key"
    identifier = CryptographicIdentifier.from_pem(public_key, controller)
    public_key_multibase = identifier.as_tuple()[1]
    test_session.add(
        RemoteIdentity(
            key_id=key_id,
            controller=controller,
            public_key=public_key_multibase,
        )
    )
    await test_session.commit()
    response = test_client.get(
        "/auth",
        headers=valid_headers_for_key_id(key_id),
    )

    assert response.status_code == 403


async def test_auth_no_signature_accept_activity_pub_is_401(test_client):
    response = test_client.get("/auth", headers={"Accept": "application/activity+json"})

    assert response.status_code == 401


async def test_auth_no_signature_accept_also_html(test_client):
    response = test_client.get(
        "/auth", headers={"Accept": "application/activity+json, text/html;q=0.5"}
    )

    assert response.status_code == 200

    should_server_header = response.headers["x-cattle-grid-should-serve"]

    assert should_server_header == "html"


async def test_auth_no_accept_header(test_client):
    response = test_client.get("/auth")

    assert response.status_code == 200

    should_server_header = response.headers["x-cattle-grid-should-serve"]

    assert should_server_header == "other"



import logging

from uuid import uuid4
from fastapi import APIRouter, HTTPException


from cattle_grid.database.account import AuthenticationToken
from cattle_grid.account.account import account_with_name_password
from cattle_grid.dependencies.fastapi import CommittingSqlSession

from .responses import TokenResponse, SignInData
from .actor import actor_router
from .account import account_router

logger = logging.getLogger(__name__)


router = APIRouter()
router.include_router(actor_router)
router.include_router(account_router)


@router.post("/signin", operation_id="signin")
async def signin(data: SignInData, session: CommittingSqlSession) -> TokenResponse:
    """Allows one to sign in to an account on cattle_grid.
    The response a token to be included using bearer authentication."""
    account = await account_with_name_password(session, data.name, data.password)
    if account is None:
        raise HTTPException(401)

    token = str(uuid4())
    session.add(AuthenticationToken(account=account, token=token))

    return TokenResponse(token=token)



import click
import asyncio
import logging
from pathlib import Path

from cattle_grid.config import load_settings
from .statistics import statistics
from .auth.__main__ import add_auth_commands
from .auth.block_cli import add_block_command
from .auth.key_cli import add_keys_command
from .actors import add_actors_to_cli_as_group
from .extensions.cli import add_extension_commands_as_group
from .account.cli import add_account_commands

from .cli import (
    filenames_for_component,
    async_api_schema_for_component,
    fastapi_for_component,
    filenames_for_openapi_components,
)

logger = logging.getLogger(__name__)


@click.group()
@click.option("--config_file", default="cattle_grid.toml")
@click.pass_context
def main(ctx, config_file):
    ctx.ensure_object(dict)
    ctx.obj["config_file"] = config_file

    try:
        ctx.obj["config"] = load_settings(config_file)
    except Exception as e:
        logger.Exception(e)


@main.command()
@click.pass_context
def stat(ctx):
    """Displays statistical information about cattle_grid"""
    asyncio.run(statistics(ctx.obj["config"]))


@main.command()
@click.option("--db_uri")
@click.option("--name")
@click.pass_context
def create_db_migration(ctx, db_uri, name):
    """Creates a database migration; run after editing models"""
    from .database import migrate

    config = ctx.obj["config"]

    if db_uri:
        config.db_uri = db_uri

    asyncio.run(migrate(config, name))


@main.command()
@click.option("--db_uri")
@click.pass_context
def upgrade_db(ctx, db_uri):
    from .database import upgrade

    config = ctx.obj["config"]

    if db_uri:
        config.db_uri = db_uri

    asyncio.run(upgrade(config))


@main.command()
@click.option("--filename", default="asyncapi.json")
@click.option(
    "--for_docs",
    is_flag=True,
    default=False,
    help="Generate schema for use in documentation",
)
@click.option(
    "--component",
    default="exchange",
    help="Restrict to a component. Currently allowed ap",
)
def async_api(filename: str, for_docs: bool, component: str):
    if for_docs:
        Path("./docs/assets/schemas").mkdir(parents=True, exist_ok=True)
        to_process = filenames_for_component
    else:
        to_process = {component: filename}

    for component, filename in to_process.items():
        schema = async_api_schema_for_component(component)

        with open(filename, "w") as fp:
            fp.write(schema)


@main.command()
@click.option("--filename", default="openapi.json")
@click.option(
    "--for_docs",
    is_flag=True,
    default=False,
    help="Generate schema for use in documentation",
)
@click.option(
    "--component",
    default=None,
    help="Restrict to a component. Currently allowed auth or ap",
)
def openapi(filename: str, for_docs: bool, component):
    import json

    if for_docs:
        Path("./docs/assets/schemas").mkdir(parents=True, exist_ok=True)
        to_process = filenames_for_openapi_components
    else:
        to_process = {component: filename}

    for component, filename in to_process.items():
        app = fastapi_for_component(component)

        with open(filename, "w") as fp:
            json.dump(app.openapi(), fp)


add_auth_commands(main)
add_block_command(main)
add_keys_command(main)
add_extension_commands_as_group(main)
add_account_commands(main)
add_actors_to_cli_as_group(main)

if __name__ == "__main__":
    main()



"""Simple transformer type extension that adds the recipients and whether the activity is public to the data. You can
include this transformer into cattle_grid by adding

```toml title="cattle_grid.toml"
[[extensions]]
module = "cattle_grid.extensions.examples.recipients"
```

to your configuration file.

"""

from cattle_grid.extensions import Extension
from bovine.activitystreams.utils import recipients_for_object, is_public


extension = Extension("recipients", __name__)


@extension.transform(inputs=["raw"], outputs=["recipients"])
async def transformer(data):
    """The transformer

    ```pycon
    >>> from asyncio import run
    >>> run(transformer({"raw": {"type": "Activity", "to": ["http://alice.example"]}}))
    {'recipients': {'recipients': ['http://alice.example'], 'public': False}}

    ```
    """
    raw = data.get("raw")
    recipients = recipients_for_object(raw)
    public = is_public(raw)

    return {"recipients": {"recipients": list(recipients), "public": public}}



import pytest

from unittest.mock import AsyncMock

from cattle_grid.model import ActivityMessage
from cattle_grid.activity_pub.actor import create_actor
from cattle_grid.database.activity_pub_actor import Blocking
from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.model.account import EventType

from .shovel import incoming_shovel, should_shovel_activity, shovel_to_account_exchange


async def fake_transformer(a):
    return a


async def test_incoming_message(sql_session_for_test, actor_with_account):
    broker = AsyncMock()

    activity_pub = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://remote.test/activity_id",
        "type": "Activity",
        "actor": "http://remote.test/actor",
        "to": "http://actor.example",
        "object": "http://actor.example/object",
    }

    activity = ActivityMessage(
        actor=actor_with_account.actor_id,
        data=activity_pub,
    )

    await incoming_shovel(
        activity, fake_transformer, "uudi", session=sql_session_for_test, broker=broker
    )

    broker.publish.assert_awaited()
    assert len(broker.publish.call_args) == 2
    args = broker.publish.call_args[0]
    result = args[0]

    assert result["data"]["raw"]["@context"] == activity_pub["@context"]


async def test_incoming_block(sql_session_for_test, actor_with_account):
    broker = AsyncMock()

    activity_pub = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://remote.test/activity_id",
        "type": "Block",
        "actor": "http://remote.test/actor",
        "to": "http://actor.example",
        "object": "http://actor.example/object",
    }

    activity = ActivityMessage(actor=actor_with_account.actor_id, data=activity_pub)

    await incoming_shovel(
        activity, fake_transformer, "uuid", session=sql_session_for_test, broker=broker
    )

    broker.publish.assert_not_awaited()


async def test_incoming_from_blocked_user(sql_session_for_test, actor_with_account):
    broker = AsyncMock()

    remote_actor = "http://remote.test/actor"

    sql_session_for_test.add(
        Blocking(
            actor=actor_with_account,
            blocking=remote_actor,
            request="http://blocked.test/id",
            active=True,
        )
    )
    await sql_session_for_test.commit()

    activity_pub = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://remote.test/activity_id",
        "type": "Activity",
        "actor": remote_actor,
        "to": "http://actor.example",
        "object": "http://actor.example/object",
    }

    activity = ActivityMessage(
        actor=actor_with_account.actor_id,
        data=activity_pub,
    )

    await incoming_shovel(
        activity, fake_transformer, "uuid", sql_session_for_test, broker=broker
    )

    broker.publish.assert_not_awaited()


async def test_incoming_from_blocked_user_inactive_block(
    sql_session_for_test, actor_with_account
):
    broker = AsyncMock()

    remote_actor = "http://remote.test/actor"

    sql_session_for_test.add(
        Blocking(
            actor=actor_with_account,
            blocking=remote_actor,
            request="http://blocked.test/id",
            active=False,
        )
    )
    await sql_session_for_test.commit()

    activity_pub = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://remote.test/activity_id",
        "type": "Activity",
        "actor": remote_actor,
        "to": "http://actor.example",
        "object": "http://actor.example/object",
    }

    activity = ActivityMessage(
        actor=actor_with_account.actor_id,
        data=activity_pub,
    )

    await incoming_shovel(
        activity, fake_transformer, "uuid", sql_session_for_test, broker=broker
    )

    broker.publish.assert_awaited()
    assert len(broker.publish.call_args) == 2


async def test_incoming_message_non_gateway_actor(sql_session_for_test):
    actor = await create_actor(
        sql_session_for_test, "http://localhost/", preferred_username="bob"
    )
    activity_pub = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://remote.test/activity_id",
        "type": "Activity",
        "actor": "http://remote.test/actor",
        "to": "http://actor.example",
        "object": "http://actor.example/object",
    }

    broker = AsyncMock()
    activity = ActivityMessage(
        actor=actor.actor_id,
        data=activity_pub,
    )

    await incoming_shovel(
        activity, fake_transformer, "id", session=sql_session_for_test, broker=broker
    )

    broker.publish.assert_not_awaited()


@pytest.mark.parametrize(
    "activity, expected",
    [
        (
            {
                "type": "Activity",
            },
            True,
        ),
        (
            {
                "type": "Block",
            },
            False,
        ),
        (
            {"type": "Undo", "object": "http://follow.test/id"},
            True,
        ),
    ],
)
async def test_should_shovel_activity(sql_session_for_test, activity, expected):
    result = await should_shovel_activity(sql_session_for_test, activity)

    assert result == expected


@pytest.mark.parametrize("active, expected", [(True, False), (False, False)])
async def test_should_shovel_activity_undo(
    sql_session_for_test, actor_with_account, active, expected
):
    block_id = "http://block.test/id"
    activity = {"type": "Undo", "object": block_id}

    sql_session_for_test.add(
        Blocking(
            actor=actor_with_account,
            blocking="http://remote.test",
            request=block_id,
            active=active,
        )
    )
    await sql_session_for_test.commit()

    result = await should_shovel_activity(sql_session_for_test, activity)

    assert result == expected


async def test_shovel_to_account_exchange():
    broker = AsyncMock()

    await shovel_to_account_exchange(
        "actor", "name", EventType.incoming, {"raw": "data"}, broker, "uuid"
    )

    broker.publish.assert_awaited()
    args, kwargs = broker.publish.call_args

    assert kwargs["routing_key"] == "receive.name.incoming"
    assert args[0].actor == "actor"



from contextlib import asynccontextmanager
from typing import Callable, List, AsyncContextManager, Any
from fast_depends import inject
from fastapi import FastAPI

from cattle_grid.dependencies.globals import global_container
from cattle_grid.extensions import Extension
from cattle_grid.model.extension import AddMethodInformationMessage


def collect_lifespans(
    extensions: List[Extension],
) -> List[Callable[[Any], AsyncContextManager]]:
    """Collects the lifespans from the extensions"""
    return [extension.lifespan for extension in extensions if extension.lifespan]


@asynccontextmanager
async def iterate_lifespans(lifespans: List[Callable[[Any], AsyncContextManager]]):
    if len(lifespans) == 0:
        yield
        return

    async with inject(lifespans[0])():  # type: ignore
        if len(lifespans) == 1:
            yield
        else:
            async with iterate_lifespans(lifespans[1:]):
                yield


def lifespan_for_extension(extension: Extension, include_broker: bool = False):
    broker = None

    if include_broker and extension.activity_router:
        broker = global_container.broker
        broker.include_router(extension.activity_router)

    @asynccontextmanager
    async def lifespan(app: FastAPI):
        async with global_container.common_lifecycle():
            if broker:
                await broker.start()

                await broker.publish(
                    AddMethodInformationMessage(
                        method_information=extension.method_information
                    ),
                    exchange=global_container.exchange,
                    routing_key="add_method_information",
                )

            if extension.lifespan:
                async with inject(extension.lifespan)():  # type: ignore
                    yield
            else:
                yield

            if broker:
                await broker.close()

    return lifespan


@asynccontextmanager
async def lifespan_from_extensions(extensions: list[Extension]):
    """Creates the lifespan from the extensions"""
    lifespans = collect_lifespans(extensions)

    async with iterate_lifespans(lifespans):
        yield
